import numpy as np
import math
from ..pyramids import convert_pyr_coeffs_to_pyr


def comparePyr(matPyr, pyPyr, rtol=1e-5, atol=1e-8):
    '''compare two pyramids

    returns True if they are the same with in desired precision and False if not.

    written for unit testing code.

    '''
    # compare two pyramids - return 0 for !=, 1 for ==
    # correct number of elements?
    matSz = sum(matPyr.shape)
    try:
        pySz = 1 + sum([np.asarray(size).prod() for size in pyPyr.pyr_size.values()])
    except AttributeError:
        pySz = 1 + sum([np.asarray(size).prod() for size in pyPyr.pyrSize])

    if(matSz != pySz):
        print("size difference: %d != %d, returning False" % (matSz, pySz))
        return False

    # values are close to each other?
    matStart = 0
    try:
        pyCoeffs, pyHigh, pyLow = convert_pyr_coeffs_to_pyr(pyPyr.pyr_coeffs)
        if pyHigh is not None:
            pyCoeffs.insert(0, pyHigh)
        if pyLow is not None:
            pyCoeffs.append(pyLow)
    except AttributeError:
        pyCoeffs = pyPyr.pyr
    for idx, pyTmp in enumerate(pyCoeffs):
        matTmp = matPyr[matStart:matStart + pyTmp.size]
        matStart = matStart + pyTmp.size
        matTmp = np.reshape(matTmp, pyTmp.shape, order='F')

        # relative tolerance rtol
        # absolute tolerance atol
        isclose = np.isclose(matTmp, pyTmp, rtol, atol)
        if not isclose.all():
            print("some pyramid elements not identical: checking...")
            for i in range(isclose.shape[0]):
                for j in range(isclose.shape[1]):
                    if not isclose[i, j]:
                        print("failed level:%d element:%d %d value:%.15f %.15f" %
                              (idx, i, j, matTmp[i, j], pyTmp[i, j]))
                        return False

    return True


def compareRecon(recon1, recon2, rtol=1e-5, atol=1e-10):
    '''compare two arrays

    returns True is they are the same within specified precision and False if not.  function was
    made to accompany unit test code.

    This function is deprecated. Instead use the builtin numpy:

    np.allclose(recon1, recon2, rtol=1e-05, atol=1e-08, equal_nan=False)

    This will not tell you where the error is, but you can find that yourself

    '''

    # NOTE builtin numpy:

    # BUT won't print where and what the first error is

    if recon1.shape != recon2.shape:
        print('shape is different!')
        print(recon1.shape)
        print(recon2.shape)
        return False

    prec = -1
    for i in range(recon1.shape[0]):
        for j in range(recon2.shape[1]):
            if np.absolute(recon1[i, j].real - recon2[i, j].real) > math.pow(10, -11):
                print("real: i=%d j=%d %.15f %.15f diff=%.15f" %
                      (i, j, recon1[i, j].real, recon2[i, j].real,
                       np.absolute(recon1[i, j].real-recon2[i, j].real)))
                return False
            # FIX: need a better way to test
            # if we have many significant digits to the left of decimal we
            # need to be less stringent about digits to the right.
            # The code below works, but there must be a better way.
            if isinstance(recon1, complex):
                if int(math.log(np.abs(recon1[i, j].imag), 10)) > 1:
                    prec = prec + int(math.log(np.abs(recon1[i, j].imag), 10))
                    if prec > 0:
                        prec = -1
                print(prec)
                if np.absolute(recon1[i, j].imag - recon2[i, j].imag) > math.pow(10, prec):
                    print("imag: i=%d j=%d %.15f %.15f diff=%.15f" %
                          (i, j, recon1[i, j].imag, recon2[i, j].imag,
                           np.absolute(recon1[i, j].imag-recon2[i, j].imag)))
                    return False

    return True



import numpy as np
from ..pyramids.filters import parse_filter
from ..pyramids.c.wrapper import corrDn, upConv
import scipy.signal


def blur(image, n_levels=1, filt='binom5'):
    '''blur an image by filtering-downsampling and then upsampling-filtering

    Blur an image, by filtering and downsampling `n_levels` times (default=1), followed by upsampling
    and filtering `n_levels` times.  The blurring is done with filter kernel specified by `filt` (default
    = 'binom5'), which can be a string (to be passed to named_filter), a vector (applied separably
    as a 1D convolution kernel in X and Y), or a matrix (applied as a 2D convolution kernel).  The
    downsampling is always by 2 in each direction.

    This differs from blurDn in that here we upsample afterwards.

    Arguments
    ---------
    image : `array_like`
        1d or 2d image to blur
    n_levels : `int`
        the number of times to filter and downsample. the higher this is, the more blurred the
        resulting image will be
    filt : {`array_like`, 'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3', 'daub4',
            'qmf5', 'qmf9', 'qmf13'}
        filter to use for filtering image. If array_like, can be 1d or 2d. All scaled so L-1 norm
        is 1.0

        * `'binomN'` - binomial coefficient filter of order N-1
        * `'haar'` - Haar wavelet
        * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
        * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
        * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_

    Returns
    -------
    image : `array_like`
        the blurred image

    References
    ----------
    .. [1] J D Johnston, "A filter family designed for use in quadrature mirror filter banks",
       Proc. ICASSP, pp 291-294, 1980.
    .. [2] I Daubechies, "Orthonormal bases of compactly supported wavelets", Commun. Pure Appl.
       Math, vol. 42, pp 909-996, 1988.
    .. [3] E P Simoncelli,  "Orthogonal sub-band image transforms", PhD Thesis, MIT Dept. of Elec.
       Eng. and Comp. Sci. May 1988. Also available as: MIT Media Laboratory Vision and Modeling
       Technical Report #100.
    .. [4] E P Simoncelli and E H Adelson, "Subband image coding", Subband Transforms, chapter 4,
       ed. John W Woods, Kluwer Academic Publishers,  Norwell, MA, 1990, pp 143--192.
    '''

    if image.ndim == 1:
        image = image.reshape(-1, 1)

    filt = parse_filter(filt)

    if n_levels > 0:
        if image.shape[1] == 1:
            # 1D image [M, 1] 1D filter [N, 1]
            imIn = corrDn(image=image, filt=filt, step=(2, 1))
            out = blur(imIn, n_levels-1, filt)
            res = upConv(image=out, filt=filt, step=(2, 1), stop=image.shape)
            return res

        elif image.shape[1] == 1:
            # 1D image [1, M] 1D filter [N, 1]
            imIn = corrDn(image=image, filt=filt.T, step=(1, 2))
            out = blur(imIn, n_levels-1, filt)
            res = upConv(image=out, filt=filt.T, step=(1, 2), stop=image.shape)
            return res

        elif filt.shape[1] == 1:
            # 2D image 1D filter [N, 1]
            imIn = corrDn(image=image, filt=filt, step=(2, 1))
            imIn = corrDn(image=imIn, filt=filt.T, step=(1, 2))
            out = blur(imIn, n_levels-1, filt)
            res = upConv(image=out, filt=filt.T, step=(1, 2), start=(0, 0), stop=[out.shape[0], image.shape[1]])
            res = upConv(image=res, filt=filt, step=(2, 1), start=(0, 0), stop=image.shape)
            return res

        else:
            # 2D image 2D filter
            imIn = corrDn(image=image, filt=filt, step=(2, 2))
            out = blur(imIn, n_levels-1, filt)
            res = upConv(image=out, filt=filt, step=(2, 2), stop=image.shape)
            return res

    else:
        return image


def blurDn(image, n_levels=1, filt='binom5'):
    '''blur and downsample an image

    Blur and downsample an image.  The blurring is done with filter kernel specified by FILT
    (default = 'binom5'), which can be a string (to be passed to named_filter), a vector (applied
    separably as a 1D convolution kernel in X and Y), or a matrix (applied as a 2D convolution
    kernel).  The downsampling is always by 2 in each direction.

    The procedure is applied recursively `n_levels` times (default=1).

    This differs from blur in that we do NOT upsample afterwards.

    Arguments
    ---------
    image : `array_like`
        1d or 2d image to blur and downsample
    n_levels : `int`
        the number of times to filter and downsample. the higher this is, the blurrier and smaller
        the resulting image will be
    filt : {`array_like`, 'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3', 'daub4',
            'qmf5', 'qmf9', 'qmf13'}
        filter to use for filtering image. If array_like, can be 1d or 2d. All scaled so L-1 norm
        is 1.0

        * `'binomN'` - binomial coefficient filter of order N-1
        * `'haar'` - Haar wavelet
        * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
        * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
        * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_

    Returns
    -------
    image : `array_like`
        the blurred and downsampled image

    References
    ----------
    .. [1] J D Johnston, "A filter family designed for use in quadrature mirror filter banks",
       Proc. ICASSP, pp 291-294, 1980.
    .. [2] I Daubechies, "Orthonormal bases of compactly supported wavelets", Commun. Pure Appl.
       Math, vol. 42, pp 909-996, 1988.
    .. [3] E P Simoncelli,  "Orthogonal sub-band image transforms", PhD Thesis, MIT Dept. of Elec.
       Eng. and Comp. Sci. May 1988. Also available as: MIT Media Laboratory Vision and Modeling
       Technical Report #100.
    .. [4] E P Simoncelli and E H Adelson, "Subband image coding", Subband Transforms, chapter 4,
       ed. John W Woods, Kluwer Academic Publishers,  Norwell, MA, 1990, pp 143--192.
    '''

    if image.ndim == 1:
        image = image.reshape(-1, 1)

    filt = parse_filter(filt)

    if n_levels > 1:
        image = blurDn(image, n_levels-1, filt)

    if n_levels >= 1:
        if image.shape[1] == 1:
            # 1D image [M, 1] and 1D filter [N, 1]
            res = corrDn(image=image, filt=filt, step=(2, 1))

        elif image.shape[0] == 1:
            # 1D image [1, M] and 1D filter [N, 1]
            res = corrDn(image=image, filt=filt.T, step=(1, 2))

        elif filt.shape[1] == 1:
            # 2D image and 1D filter [N, 1]
            res = corrDn(image=image, filt=filt, step=(2, 1))
            res = corrDn(image=res, filt=filt.T, step=(1, 2))

        else:
            # 2D image and 2D filter
            res = corrDn(image=image, filt=filt, step=(2, 2))

    else:
        res = image

    return res


def upBlur(image, n_levels=1, filt='binom5'):
    '''upsample and blur an image.

    Upsample and blur an image.  The blurring is done with filter kernel specified by FILT (default
    = 'binom5'), which can be a string (to be passed to named_filter), a vector (applied separably
    as a 1D convolution kernel in X and Y), or a matrix (applied as a 2D convolution kernel).  The
    downsampling is always by 2 in each direction.

    The procedure is applied recursively n_levels times (default=1).

    Arguments
    ---------
    image : `array_like`
        1d or 2d image to upsample and blur
    n_levels : `int`
        the number of times to filter and downsample. the higher this is, the blurrier and larger
        the resulting image will be
    filt : {`array_like`, 'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3', 'daub4',
            'qmf5', 'qmf9', 'qmf13'}
        filter to use for filtering image. If array_like, can be 1d or 2d. All scaled so L-1 norm
        is 1.0

        * `'binomN'` - binomial coefficient filter of order N-1
        * `'haar'` - Haar wavelet
        * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
        * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
        * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_

    Returns
    -------
    image : `array_like`
        the upsampled and blurred image

    References
    ----------
    .. [1] J D Johnston, "A filter family designed for use in quadrature mirror filter banks",
       Proc. ICASSP, pp 291-294, 1980.
    .. [2] I Daubechies, "Orthonormal bases of compactly supported wavelets", Commun. Pure Appl.
       Math, vol. 42, pp 909-996, 1988.
    .. [3] E P Simoncelli,  "Orthogonal sub-band image transforms", PhD Thesis, MIT Dept. of Elec.
       Eng. and Comp. Sci. May 1988. Also available as: MIT Media Laboratory Vision and Modeling
       Technical Report #100.
    .. [4] E P Simoncelli and E H Adelson, "Subband image coding", Subband Transforms, chapter 4,
       ed. John W Woods, Kluwer Academic Publishers,  Norwell, MA, 1990, pp 143--192.

    '''

    if image.ndim == 1:
        image = image.reshape(-1, 1)

    filt = parse_filter(filt, normalize=False)

    if n_levels > 1:
        image = upBlur(image, n_levels-1, filt)

    if n_levels >= 1:
        if image.shape[1] == 1:
            # 1D image [M, 1] and 1D filter [N, 1]
            res = upConv(image=image, filt=filt, step=(2, 1))

        elif image.shape[0] == 1:
            # 1D image [1, M] and 1D filter [N, 1]
            res = upConv(image=image, filt=filt.T, step=(1, 2))

        elif filt.shape[1] == 1:
            # 2D image and 1D filter [N, 1]
            res = upConv(image=image, filt=filt, step=(2, 1))
            res = upConv(image=res, filt=filt.T, step=(1, 2))

        else:
            # 2D image and 2D filter
            res = upConv(image=image, filt=filt, step=(2, 2))

    else:
        res = image

    return res


def image_gradient(image, edge_type="dont-compute"):
    '''Compute the gradient of the image using smooth derivative filters

    Compute the gradient of the image using smooth derivative filters optimized for accurate
    direction estimation.  Coordinate system corresponds to standard pixel indexing: X axis points
    rightward.  Y axis points downward.  `edges` specify boundary handling.

    Notes
    -----
    original filters from Int'l Conf Image Processing, 1994.
    updated filters 10/2003: see Farid & Simoncelli, IEEE Trans Image
                             Processing, 13(4):496-508, April 2004.

    Arguments
    ---------
    image : `array_like`
        2d array to compute the gradients of
    edge_type : {'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend', 'dont-compute'}
        Specifies how to handle edges. Options are:

        * `'circular'` - circular convolution
        * `'reflect1'` - reflect about the edge pixels
        * `'reflect2'` - reflect, doubling the edge pixels
        * `'repeat'` - repeat the edge pixels
        * `'zero'` - assume values of zero outside image boundary
        * `'extend'` - reflect and invert
        * `'dont-compute'` - zero output when filter overhangs imput boundaries.

    Returns
    -------
    dx, dy : `np.array`
        the X derivative and the Y derivative

    '''

    # kernels from Farid & Simoncelli, IEEE Trans Image Processing,
    #   13(4):496-508, April 2004.
    gp = np.asarray([0.037659,  0.249153, 0.426375, 0.249153, 0.037659]).reshape(5, 1)
    gd = np.asarray([-0.109604, -0.276691, 0.000000, 0.276691, 0.109604]).reshape(5, 1)

    dx = corrDn(corrDn(image, gp, edge_type), gd.T, edge_type)
    dy = corrDn(corrDn(image, gd, edge_type), gp.T, edge_type)

    return (dx, dy)


# ----------------------------------------------------------------
# Below are (slow) scipy convolution functions
# they are intended for comparison purpose only
# the c code is prefered and used throughout this package
# ----------------------------------------------------------------


def rconv2(mtx1, mtx2, ctr=0):
    '''Convolution of two matrices, with boundaries handled via reflection about the edge pixels.

    Result will be of size of LARGER matrix.

    The origin of the smaller matrix is assumed to be its center.
    For even dimensions, the origin is determined by the CTR (optional)
    argument:
         CTR   origin
          0     DIM/2      (default)
          1   (DIM/2)+1

    In general, you should not use this function, since it will be slow. Instead, use `upConv` or
    `corrDn`, which use the C code and so are much faster.

    '''

    if (mtx1.shape[0] >= mtx2.shape[0] and mtx1.shape[1] >= mtx2.shape[1]):
        large = mtx1
        small = mtx2
    elif (mtx1.shape[0] <= mtx2.shape[0] and mtx1.shape[1] <= mtx2.shape[1]):
        large = mtx2
        small = mtx1
    else:
        print('one matrix must be larger than the other in both dimensions!')
        return

    ly = large.shape[0]
    lx = large.shape[1]
    sy = small.shape[0]
    sx = small.shape[1]

    # These values are one less than the index of the small mtx that falls on
    # the border pixel of the large matrix when computing the first
    # convolution response sample:
    sy2 = int(np.floor((sy+ctr-1)/2))
    sx2 = int(np.floor((sx+ctr-1)/2))

    # pad with reflected copies
    nw = large[sy-sy2-1:0:-1, sx-sx2-1:0:-1]
    n = large[sy-sy2-1:0:-1, :]
    ne = large[sy-sy2-1:0:-1, lx-2:lx-sx2-2:-1]
    w = large[:, sx-sx2-1:0:-1]
    e = large[:, lx-2:lx-sx2-2:-1]
    sw = large[ly-2:ly-sy2-2:-1, sx-sx2-1:0:-1]
    s = large[ly-2:ly-sy2-2:-1, :]
    se = large[ly-2:ly-sy2-2:-1, lx-2:lx-sx2-2:-1]

    n = np.column_stack((nw, n, ne))
    c = np.column_stack((w, large, e))
    s = np.column_stack((sw, s, se))

    clarge = np.concatenate((n, c), axis=0)
    clarge = np.concatenate((clarge, s), axis=0)

    return scipy.signal.convolve(clarge, small, 'valid')


# TODO: low priority

# def cconv2(mtx1, mtx2, ctr=0):
#     '''Circular convolution of two matrices.  Result will be of size of
#     LARGER vector.
#
#     The origin of the smaller matrix is assumed to be its center.
#     For even dimensions, the origin is determined by the CTR (optional)
#     argument:
#          CTR   origin
#           0     DIM/2      (default)
#           1     (DIM/2)+1
#
#     Eero Simoncelli, 6/96.  Modified 2/97.
#     Python port by Rob Young, 8/15
#     '''
#
#     if len(args) < 2:
#         print 'Error: cconv2 requires two input matrices!'
#         print 'Usage: cconv2(matrix1, matrix2, center)'
#         print 'where center parameter is optional'
#         return
#     else:
#         a = np.asarray(args[0])
#         b = np.asarray(args[1])
#
#     if len(args) == 3:
#         ctr = args[2]
#     else:
#         ctr = 0
#
#     if a.shape[0] >= b.shape[0] and a.shape[1] >= b.shape[1]:
#         large = a
#         small = b
#     elif a.shape[0] <= b.shape[0] and a.shape[1] <= b.shape[1]:
#         large = b
#         small = a
#     else:
#         print 'Error: one matrix must be larger than the other in both dimensions!'
#         return
#
#     ly = large.shape[0]
#     lx = large.shape[1]
#     sy = small.shape[0]
#     sx = small.shape[1]
#
#     ## These values are the index of the small mtx that falls on the
#     ## border pixel of the large matrix when computing the first
#     ## convolution response sample:
#     sy2 = np.floor((sy+ctr+1)/2.0).astype(int)
#     sx2 = np.floor((sx+ctr+1)/2.0).astype(int)
#
#     # pad
#     nw = large[ly-sy+sy2:ly, lx-sx+sx2:lx]
#     n = large[ly-sy+sy2:ly, :]
#     ne = large[ly-sy+sy2:ly, :sx2-1]
#     w = large[:, lx-sx+sx2:lx]
#     c = large
#     e = large[:, :sx2-1]
#     sw = large[:sy2-1, lx-sx+sx2:lx]
#     s = large[:sy2-1, :]
#     se = large[:sy2-1, :sx2-1]
#
#     n = np.column_stack((nw, n, ne))
#     c = np.column_stack((w,large,e))
#     s = np.column_stack((sw, s, se))
#
#     clarge = np.concatenate((n, c), axis=0)
#     clarge = np.concatenate((clarge, s), axis=0)
#
#     c = scipy.signal.convolve(clarge, small, 'valid')
#
#     return c


# def zconv2(mtx1, mtx2, ctr=0):
#     ''' RES = ZCONV2(MTX1, MTX2, CTR)
#
#         Convolution of two matrices, with boundaries handled as if the larger
#         mtx lies in a sea of zeros. Result will be of size of LARGER vector.
#
#         The origin of the smaller matrix is assumed to be its center.
#         For even dimensions, the origin is determined by the CTR (optional)
#         argument:
#              CTR   origin
#               0     DIM/2      (default)
#               1     (DIM/2)+1  (behaves like conv2(mtx1,mtx2,'same'))
#
#         Eero Simoncelli, 2/97.  Python port by Rob Young, 10/15.  '''
#
#     # REQUIRED ARGUMENTS
#     #----------------------------------------------------------------
#
#     if len(args) < 2 or len(args) > 3:
#         print 'Usage: zconv2(matrix1, matrix2, center)'
#         print 'first two input parameters are required'
#         return
#     else:
#         a = np.asarray(args[0])
#         b = np.asarray(args[1])
#
#     # OPTIONAL ARGUMENT
#     #----------------------------------------------------------------
#
#     if len(args) == 3:
#         ctr = args[2]
#     else:
#         ctr = 0
#
#     #----------------------------------------------------------------
#
#     if (a.shape[0] >= b.shape[0]) and (a.shape[1] >= b.shape[1]):
#         large = a
#         small = b
#     elif (a.shape[0] <= b.shape[0]) and (a.shape[1] <= b.shape[1]):
#         large = b
#         small = a
#     else:
#         print 'Error: one arg must be larger than the other in both dimensions!'
#         return
#
#     ly = large.shape[0]
#     lx = large.shape[1]
#     sy = small.shape[0]
#     sx = small.shape[1]
#
#     ## These values are the index of the small matrix that falls on the
#     ## border pixel of the large matrix when computing the first
#     ## convolution response sample:
#     sy2 = np.floor((sy+ctr+1)/2.0).astype(int)-1
#     sx2 = np.floor((sx+ctr+1)/2.0).astype(int)-1
#
#     clarge = scipy.signal.convolve(large, small, 'full')
#
#     c = clarge[sy2:ly+sy2, sx2:lx+sx2]
#
#     return c



import numpy as np
from scipy import ndimage
import warnings


def matlab_round(array):
    '''round equivalent to matlab function, which rounds .5 away from zero

    (used in matlab_histo so we can unit test against matlab code).

    But np.round() would rounds .5 to nearest even number, e.g.:
    - np.round(0.5) = 0, matlab_round(0.5) = 1
    - np.round(2.5) = 2, matlab_round(2.5) = 3

    Arguments
    ---------
    array : `np.array`
        the array to round

    Returns
    -------
    rounded_array : `np.array`
        the rounded array
    '''
    (fracPart, intPart) = np.modf(array)
    return intPart + (np.abs(fracPart) >= 0.5) * np.sign(fracPart)


def matlab_histo(array, nbins=101, binsize=None, center=None):
    '''Compute a histogram of array.

    How does this differ from MatLab's HIST function?  This function:
      - allows uniformly spaced bins only.
      + operates on all elements of MTX, instead of columnwise.
      + is much faster (approximately a factor of 80 on my machine).
      + allows specification of number of bins OR binsize. Default=101 bins.
      + allows (optional) specification of binCenter.

    Arguments
    ---------
    array : `np.array`
        the array to bin
    nbins : `int`
        the number of histogram bins
    binsize : `float` or None
        the size of each bin. if None, we use nbins to determine it as:
        `(array.max() - array.min()) / nbins`
    center : `float` or None
        the center position for the histogram bins. if None, this is `array.mean()`

    Returns
    -------
    N : `np.array`
        the histogram counts
    edges : `np.array`
        vector containing the centers of the histogram bins
    '''
    mini = array.min()
    maxi = array.max()

    if center is None:
        center = array.mean()

    if binsize is None:
        # use nbins to determine binsize
        binsize = (maxi-mini) / nbins

    nbins2 = int(matlab_round((maxi - center) / binsize) - matlab_round((mini - center) / binsize))
    if nbins2 != nbins:
        warnings.warn('Overriding bin number %d (requested %d)' % (nbins2, nbins))
        nbins = nbins2

    # np.histogram uses bin edges, not centers like Matlab's hist
    # compute bin edges (nbins + 1 of them)
    edge_left = center + binsize * (-0.499 + matlab_round((mini - center) / binsize))
    edges = edge_left + binsize * np.arange(nbins+1)
    N, _ = np.histogram(array, edges)

    # matlab version returns column vectors, so we will too.
    # to check: return edges or centers? edit comments.
    return (N.reshape(1, -1), edges.reshape(1, -1))


def rcosFn(width=1, position=0, values=(0, 1)):
    '''Return a lookup table containing a "raised cosine" soft threshold function

    Y =  VALUES(1) + (VALUES(2)-VALUES(1)) * cos^2( PI/2 * (X - POSITION + WIDTH)/WIDTH )

    this lookup table is suitable for use by `pointOp`

    Arguments
    ---------
    width : `float`
        the width of the region over which the transition occurs
    position : `float`
        the location of the center of the threshold
    values : `tuple`
        2-tuple specifying the values to the left and right of the transition.

    Returns
    -------
    X : `np.array`
        the x valuesof this raised cosine
    Y : `np.array`
        the y valuesof this raised cosine
    '''

    sz = 256   # arbitrary!

    X = np.pi * np.arange(-sz-1, 2) / (2*sz)

    Y = values[0] + (values[1]-values[0]) * np.cos(X)**2

    # make sure end values are repeated, for extrapolation...
    Y[0] = Y[1]
    Y[sz+2] = Y[sz+1]

    X = position + (2*width/np.pi) * (X + np.pi/4)

    return (X, Y)


def project_polar_to_cartesian(data):
    """Take a function defined in polar coordinates and project it into Cartesian coordinates

    Inspired by https://pyabel.readthedocs.io/en/latest/_modules/abel/tools/polar.html, which went
    the other way. Note that we currently don't implement the Cartesian to polar projection, but
    could do so based on this code fairly simply if it's necessary.

    Currently, this only works for square images and we require that the original image and the
    reprojected image are the same size. There should be a way to avoid both of these issues, but I
    can't think of a way to do that right now.

    Parameters
    ----------
    data : array_like
        The 2d array to convert from polar to Cartesian coordinates. We assume the first dimension
        is the polar radius and the second is the polar angle.

    Returns
    -------
    output : np.array
        The 2d array in Cartesian coordinates.

    """
    if np.isnan(data).any():
        data[np.isnan(data)] = 0
        warnings.warn("project_polar_to_cartesian won't work if there are any NaNs in the array, "
                      "so we've replaced all NaNs with 0s")
    nx = data.shape[1]
    ny = data.shape[0]
    if nx != ny:
        raise Exception("There's an occasional bug where we don't wrap the angle correctly if nx "
                        "and ny aren't equal, so we don't support this for now!")

    max_radius = data.shape[0]
    x_i = np.linspace(-max_radius, max_radius, nx, endpoint=False)
    y_i = np.linspace(-max_radius, max_radius, ny, endpoint=False)
    x_grid, y_grid = np.meshgrid(x_i, y_i)
    # need to flip the y indices so that negative is at the bottom (to correspond with how we have
    # the polar angle -- 0 on the right)
    y_grid = np.flipud(y_grid)

    r = np.sqrt(x_grid**2 + y_grid**2)

    theta = np.arctan2(y_grid, x_grid)
    # having the angle run from 0 to 2 pi seems to avoid most of the discontinuities
    theta = np.mod(theta, 2*np.pi)
    # need to convert from 2pi to pixel values
    theta *= nx/(2*np.pi)

    r_i, theta_i = r.flatten(), theta.flatten()
    # map_coordinates requires a 2xn array
    coords = np.vstack((r_i, theta_i))
    # we use mode="nearest" to deal with weird discontinuities that may pop up near the theta=0
    # line
    zi = ndimage.map_coordinates(data, coords, mode='nearest')
    output = zi.reshape((ny, nx))
    return output


if __name__ == "__main__":
    X, Y = rcosFn(width=1, position=0, values=(0, 1))

    import matplotlib.pyplot as plt
    plt.figure()
    plt.plot(X, Y)
    plt.show()



import numpy as np
from ..pyramids.c.wrapper import pointOp
from .utils import rcosFn
from .image_stats import var


def ramp(size, direction=0, slope=1, intercept=0, origin=None):
    '''make a ramp matrix

    Compute a matrix containing samples of a ramp function in a given direction.

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the ramp should be of dimensions `(size, size)`. if a tuple, must be a
        2-tuple of ints specifying the dimensions
    direction : `float`
        the direction of the ramp's gradient direction, in radians, clockwise from the X-axis
    slope : `float`
        the slope of the ramp (per pixel)
    intercept : `intercept`
        the value of the ramp at the origin
    origin : `int`, `tuple`, or None
        the origin of the matrix. if an int, we assume the origin is at `(origin, origin)`. if a
        tuple, must be a 2-tuple of ints specifying the origin (where `(0, 0)` is the upper left).
        if None, we assume the origin lies at the center of the matrix, `(size-1)/2`.

    Returns
    -------
    res : `np.array`
        the ramp matrix
    '''
    if not hasattr(size, '__iter__'):
        size = (size, size)

    if origin is None:
        # TODO understand why minus one (not plus)
        origin = ((size[0] - 1)/2., (size[1] - 1)/2.)
        # origin = ( (size[0] + 1)/2., (size[1] + 1)/2. )
    elif not hasattr(origin, '__iter__'):
        origin = (origin, origin)

    xinc = slope * np.cos(direction)
    yinc = slope * np.sin(direction)

    [xramp, yramp] = np.meshgrid(xinc * (np.arange(size[1])-origin[1]),
                                 yinc * (np.arange(size[0])-origin[0]))

    res = intercept + xramp + yramp

    return res


def impulse(size, origin=None, amplitude=1):
    '''make an impulse matrix

    create an image that is all zeros except for an impulse of a given amplitude at the origin

    NOTE: the origin is rounded to the nearest int

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size, size)`. if a tuple, must
        be a 2-tuple of ints specifying the dimensions
    origin : `int`, `tuple`, or None
        the location of the impulse. if an int, we assume the origin is at `(origin, origin)`. if a
        tuple, must be a 2-tuple of ints specifying the origin (where `(0, 0)` is the upper left).
        if None, we assume the origin lies at the center of the matrix, `(size+1)//2` (note: this
        is rounded to the nearest int)
    amplitude : `float`
        the amplitude of the impulse

    Returns
    -------
    res : `np.array`
        the impulse matrix
    '''

    if not hasattr(size, '__iter__'):
        size = (size, size)

    if origin is None:
        origin = ((size[0] + 1)//2, (size[1] + 1)//2)
    elif not hasattr(origin, '__iter__'):
        origin = (origin, origin)

    res = np.zeros(size)
    res[origin[0], origin[1]] = amplitude

    return res


def polar_radius(size, exponent=1, origin=None):
    '''make distance-from-origin (r) matrix

    Compute a matrix of given size containing samples of a radial ramp function, raised to given
    exponent, centered at given origin.

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size, size)`. if a tuple, must be
        a 2-tuple of ints specifying the dimensions
    exponent : `float`
        the exponent of the radial ramp function.
    origin : `int`, `tuple`, or None
        the center of the image. if an int, we assume the origin is at `(origin, origin)`. if a
        tuple, must be a 2-tuple of ints specifying the origin (where `(0, 0)` is the upper left).
        if None, we assume the origin lies at the center of the matrix, `(size+1)/2`.

    Returns
    -------
    res : `np.array`
        the polar radius matrix

    '''
    if not hasattr(size, '__iter__'):
        size = (size, size)

    if origin is None:
        origin = ((size[0]+1)/2., (size[1]+1)/2.)
    elif not hasattr(origin, '__iter__'):
        origin = (origin, origin)

    xramp, yramp = np.meshgrid(np.arange(1, size[1]+1)-origin[1],
                               np.arange(1, size[0]+1)-origin[0])

    if exponent <= 0:
        # zero to a negative exponent raises:
        # ZeroDivisionError: 0.0 cannot be raised to a negative power
        r = xramp ** 2 + yramp ** 2
        res = np.power(r, exponent / 2.0, where=(r != 0))
    else:
        res = (xramp ** 2 + yramp ** 2) ** (exponent / 2.0)
    return res


def polar_angle(size, phase=0, origin=None, direction='clockwise'):
    '''make polar angle matrix (in radians)

    Compute a matrix of given size containing samples of the polar angle (in radians,
    increasing in user-defined direction from the X-axis, ranging from -pi to pi), relative to
    given phase, about the given origin pixel.

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size, size)`. if a tuple, must be
        a 2-tuple of ints specifying the dimensions
    phase : `float`
        the phase of the polar angle function (in radians, clockwise from the X-axis)
    origin : `int`, `tuple`, or None
        the center of the image. if an int, we assume the origin is at `(origin, origin)`. if a
        tuple, must be a 2-tuple of ints specifying the origin (where `(0, 0)` is the upper left).
        if None, we assume the origin lies at the center of the matrix, `(size+1)/2`.
    direction : {'clockwise', 'counter-clockwise'}
        Whether the angle increases in a clockwise or counter-clockwise direction from
        the x-axis. The standard mathematical convention is to increase
        counter-clockwise, so that 90 degrees corresponds to the positive y-axis.

    Returns
    -------
    res : `np.array`
        the polar angle matrix

    '''
    if direction not in ['clockwise', 'counter-clockwise']:
        raise ValueError("direction must be one of {'clockwise', 'counter-clockwise'}, "
                         f"but received {direction}!")
    if not hasattr(size, '__iter__'):
        size = (size, size)

    if origin is None:
        origin = ((size[0]+1)/2., (size[1]+1)/2.)
    elif not hasattr(origin, '__iter__'):
        origin = (origin, origin)

    xramp, yramp = np.meshgrid(np.arange(1, size[1]+1)-origin[1],
                               np.arange(1, size[0]+1)-origin[0])
    if direction == 'counter-clockwise':
        yramp = np.flip(yramp, 0)

    res = np.arctan2(yramp, xramp)

    res = ((res+(np.pi-phase)) % (2*np.pi)) - np.pi

    return res


def disk(size, radius=None, origin=None, twidth=2, vals=(1, 0)):
    '''make a "disk" image

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size, size)`. if a tuple, must be
        a 2-tuple of ints specifying the dimensions
    radius : `float` or None
        the radius of the disk (in pixels). If None, defaults to `min(size)/4`.
    origin : `int`, `tuple`, or None
        the center of the image. if an int, we assume the origin is at `(origin, origin)`. if a
        tuple, must be a 2-tuple of ints specifying the origin (where `(0, 0)` is the upper left).
        if None, we assume the origin lies at the center of the matrix, `(size+1)/2`.
    twidth : `float`
        the width (in pixels) over which a soft threshold transition is made.
    vals : `tuple`
        2-tuple of floats containing the intensity value inside and outside the disk.

    Returns
    -------
    res : `np.array`
        the disk image matrix

    '''
    if not hasattr(size, '__iter__'):
        size = (size, size)

    if radius is None:
        radius = min(size) / 4.0

    if origin is None:
        origin = ((size[0]+1)/2., (size[1]+1)/2.)
    elif not hasattr(origin, '__iter__'):
        origin = (origin, origin)

    res = polar_radius(size, exponent=1, origin=origin)

    if abs(twidth) < np.finfo(np.double).tiny:
        res = vals[1] + (vals[0] - vals[1]) * (res <= radius)
    else:
        [Xtbl, Ytbl] = rcosFn(twidth, radius, [vals[0], vals[1]])
        res = pointOp(res, Ytbl, Xtbl[0], Xtbl[1]-Xtbl[0])

    return np.asarray(res)


def gaussian(size, covariance=None, origin=None, amplitude='norm'):
    '''make a two dimensional Gaussian

    make a two dimensional Gaussian function with specified `size`, centered at a given pixel
    position, with given covariance and amplitude

    TODO - use built in scipy function

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size, size)`. if a tuple, must be
        a 2-tuple of ints specifying the dimensions
    covariance : `float`, `np.array`, or None
        the covariance of the Gaussian. If a `float`, the covariance is [[covar, 0], [0, covar]].
        If an array, must either be of shape (2,) (e.g., [1,2]) or (2,2) (e.g., [[1,0],[0,1]]). If
        it's of shape (2,), we use [[covar[0], 0], [0, covar[1]]]. If it's of shape (2,2), we use
        it as is. If None, defaults to `(min(size)/6)^2`
    origin : `int`, `tuple`, or None
        the center of the Gaussian. if an int, we assume the origin is at `(origin, origin)`. if a
        tuple, must be a 2-tuple of ints specifying the origin (where `(0, 0)` is the upper left).
        if None, we assume the origin lies at the center of the matrix, `(size+1)/2`.
    amplitude : `float` or 'norm'
        the amplitude of the Gaussian. If 'norm', will return the probability-normalized Gaussian

    Returns
    -------
    res : `np.array`
        the 2d Gaussian

    '''
    if not hasattr(size, '__iter__'):
        size = (size, size)

    if covariance is None:
        covariance = (min([size[0], size[1]]) / 6.0) ** 2
    covariance = np.asarray(covariance)

    if origin is None:
        origin = ((size[0]+1)/2., (size[1]+1)/2.)
    elif not hasattr(origin, '__iter__'):
        origin = (origin, origin)

    (xramp, yramp) = np.meshgrid(np.arange(1, size[1]+1)-origin[1],
                                 np.arange(1, size[0]+1)-origin[0])

    if len(covariance.shape) == 0:
        if isinstance(amplitude, str) and amplitude == 'norm':
            amplitude = 1.0 / (2.0 * np.pi * covariance)
        e = ((xramp ** 2) + (yramp ** 2)) / (-2.0 * covariance)

    elif len(covariance.shape) == 1:
        if isinstance(amplitude, str) and amplitude == 'norm':
            if covariance[0] * covariance[1] < 0:
                amplitude = 1.0 / (2.0 * np.pi * np.sqrt(complex(covariance[0] * covariance[1])))
            else:
                amplitude = 1.0 / (2.0 * np.pi * np.sqrt(covariance[0] * covariance[1]))
        e = (((xramp ** 2) / (-2 * covariance[1])) +
             ((yramp ** 2) / (-2 * covariance[0])))

    elif covariance.shape == (2, 2):
        # square matrix
        if isinstance(amplitude, str) and amplitude == 'norm':
            detCov = np.linalg.det(covariance)
            if (detCov < 0).any():
                detCovComplex = np.empty(detCov.shape, dtype=complex)
                detCovComplex.real = detCov
                detCovComplex.imag = np.zeros(detCov.shape)
                amplitude = 1.0 / (2.0 * np.pi * np.sqrt(detCovComplex))
            else:
                amplitude = 1.0 / (2.0 * np.pi * np.sqrt(np.linalg.det(covariance)))
        covariance = - np.linalg.inv(covariance) / 2.0
        e = ((covariance[1, 1] * xramp**2) +
             ((covariance[0, 1] + covariance[1, 0]) * (xramp*yramp)) +
             (covariance[0, 0] * yramp**2))
    else:
        raise Exception("ERROR: invalid covariance shape")

    res = amplitude * np.exp(e)

    return res


def zone_plate(size, amplitude=1, phase=0):
    '''make a "zone plate" image

    zone plate is `amplitude` * cos( r^2 + `phase`)

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size, size)`. if a tuple, must be a
        2-tuple of ints specifying the dimensions
    amplitude : `float`
        the amplitude of the zone plate
    phase : `float`
        the phase of the zone plate (in radians, clockwise from the X-axis).

    Returns
    -------
    res : `np.array`
        the zone plate
    '''
    if not hasattr(size, '__iter__'):
        size = (size, size)

    res = amplitude * np.cos((np.pi / max(size)) * polar_radius(size, 2) + phase)

    return res


def angular_sine(size, harmonic=1, amplitude=1, phase=0, origin=None):
    '''make an angular sinusoidal image:

    the angular sinusoid is `amplitude` * sin(`harmonic`*theta + `phase`), where theta is the angle
    about the origin (clockwise from the X-axis).

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size, size)`. if a tuple, must be
        a 2-tuple of ints specifying the dimensions
    harmonic : `float`
        the frequency of the angular sinusoid.
    amplitude : `float`
        the amplitude of the angular sinusoid.
    phase : `float`
        the phase of the angular sinusoid. (in radians, clockwise from the X-axis).
    origin : `int`, `tuple`, or None
        the center of the image. if an int, we assume the origin is at `(origin, origin)`. if a
        tuple, must be a 2-tuple of ints specifying the origin (where `(0, 0)` is the upper left).
        if None, we assume the origin lies at the center of the matrix, `(size+1)/2`.

    Returns
    -------
    res : `np.array`
        the angular sinusoid

    '''
    if not hasattr(size, '__iter__'):
        size = (size, size)

    if origin is None:
        origin = ((size[0]+1)/2., (size[1]+1)/2.)
    elif not hasattr(origin, '__iter__'):
        origin = (origin, origin)

    res = amplitude * np.sin(harmonic * polar_angle(size, phase, origin) + phase)

    return res


def sine(size, period=None, direction=None, frequency=None, amplitude=1,
         phase=0, origin=None):
    '''make a two dimensional sinusoid

    this uses either the period and direction or a 2-tuple of frequencies. So either frequency or
    period and direction should be None, and the other should be set. If period is set, it takes
    precedence over frequency. If neither are set, we default to (2*pi) / (log2(size[0])).

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size, size)`. if a tuple, must be
        a 2-tuple of ints specifying the dimensions
    period : `float` or None
        the period of the two-dimensional sinusoid in pixels. If both `period` and `frequency` are
        None, we set `frequency` to (2*pi) / (log2(size[0])). If `period` is set, we ignore
        `frequency`
    direction : `float` or None
        the direction of the two-dimensional sinusoid, in radians, clockwise from the X-axis. If
        `period` is set and this is None, set to 0.
    frequency : `tuple` or None
        (f_x, f_y), the x and y frequency of the sinusoid in cycles per pixel. If both `period`
        and `frequency` are None, we set `frequency` to (2*pi) / (log2(size[0])).
    amplitude : `float`
        the amplitude of the sinusoid.
    phase : `float`
        the phase of the sinusoid (in radians, clockwise from the X-axis).
    origin : `int`, `tuple`, or None
        the center of the image. if an int, we assume the origin is at `(origin, origin)`. if a
        tuple, must be a 2-tuple of ints specifying the origin (where `(0, 0)` is the upper left).
        if None, we assume the origin lies at the center of the matrix, `(size+1)/2`.

    Returns
    -------
    res : `np.array`
        the two-dimensional sinusoid

    '''
    if not hasattr(size, '__iter__'):
        size = (size, size)

    # first form
    if isinstance(period, (int, float)):
        frequency = (2.0 * np.pi) / period
        if direction is None:
            direction = 0

    # second form
    elif frequency is not None:
        direction = np.arctan2(frequency[0], frequency[1])
        frequency = np.linalg.norm(frequency)

    elif period is None and frequency is None:
        frequency = (2.0 * np.pi) / np.log2(size[0])
        direction = 0

    if origin is None:
        res = amplitude * np.sin(ramp(size=size, direction=direction,
                                      slope=frequency, intercept=phase))

    else:
        if not hasattr(origin, '__iter__'):
            origin = (origin, origin)
        res = amplitude * np.sin(ramp(size=size, direction=direction,
                                      slope=frequency, intercept=phase,
                                      origin=[origin[0]-1, origin[1]-1]))

    return res


def square_wave(size, period=None, direction=None, frequency=None, amplitude=1,
                phase=0, origin=None, twidth=None):
    '''make a two dimensional square wave

    this uses either the period and direction or a 2-tuple of frequencies. So either frequency or
    period and direction should be None, and the other should be set. If period is set, it takes
    precedence over frequency. If neither are set, we default to (2*pi) / (log2(size[0])).

    TODO: Add duty cycle

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size, size)`. if a tuple, must be
        a 2-tuple of ints specifying the dimensions
    period : `float` or None
        the period of the square wave in pixels. If both `period` and `frequency` are None, we set
        `frequency` to (2*pi) / (log2(size[0])). If `period` is set, we ignore
        `frequency`
    direction : `float` or None
        the direction of the square wave, in radians, clockwise from the X-axis. If `period` is set
        and this is None, set to 0.
    frequency : `tuple` or None
        (f_x, f_y), the x and y frequency of the square wave in cycles per pixel. If both `period`
        and `frequency` are None, we set `frequency` to (2*pi) / (log2(size[0])).
    amplitude : `float`
        the amplitude of the sinusoid.
    phase : `float`
        the phase of the square wave (in radians, clockwise from the X-axis).
    origin : `int`, `tuple`, or None
        the center of the image. if an int, we assume the origin is at `(origin, origin)`. if a
        tuple, must be a 2-tuple of ints specifying the origin (where `(0, 0)` is the upper left).
        if None, we assume the origin lies at the center of the matrix, `(size+1)/2`.
    twidth : `float` or None
        the width of the raised-cosine edges on the bars of the grating. If None, default to
        min(2, period/3)

    Returns
    -------
    res : `np.array`
        the two-dimensional square wave
    '''
    if not hasattr(size, '__iter__'):
        size = (size, size)

    # first form
    if isinstance(period, (int, float)):
        frequency = (2.0 * np.pi) / period
        if direction is None:
            direction = 0.0

    # second form
    elif frequency is not None:
        direction = np.arctan2(frequency[0], frequency[1])
        frequency = np.linalg.norm(frequency)

    elif period is None and frequency is None:
        frequency = (2.0 * np.pi) / np.log2(size[0])
        direction = 0

    if twidth is None:
        twidth = min(2, 2.0 * np.pi / (3.0*frequency))

    if origin is None:
        res = ramp(size, direction=direction, slope=frequency,
                   intercept=phase) - np.pi/2.0

    else:
        if not hasattr(origin, '__iter__'):
            origin = (origin, origin)
        res = ramp(size, direction=direction, slope=frequency,
                   intercept=phase, origin=[origin[0]-1, origin[1]-1]) - np.pi/2.0

    [Xtbl, Ytbl] = rcosFn(twidth * frequency, np.pi/2.0,
                          [-amplitude, amplitude])

    res = pointOp(abs(((res+np.pi) % (2.0*np.pi))-np.pi), Ytbl,
                  Xtbl[0], Xtbl[1]-Xtbl[0])

    return res


def pink_noise(size, fract_dim=1):
    '''make pink noise

    Make a matrix of specified size containing fractal (pink) noise with power spectral density of
    the form: 1/f^(5-2*`fract_dim`).  Image variance is normalized to 1.0.

    TODO: Verify that this  matches Mandelbrot defn of fractal dimension.
          Make this more efficient!

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size, size)`. if a tuple, must be
        a 2-tuple of ints specifying the dimensions
    fract_dim : `float`
        the fractal dimension of the pink noise

    Returns
    -------
    res : `np.array`
        the pink noise

    '''
    if not hasattr(size, '__iter__'):
        size = (size, size)

    res = np.random.randn(size[0], size[1])
    fres = np.fft.fft2(res)

    exp = -(2.5-fract_dim)
    ctr = np.ceil((res.shape + np.ones(2))/2.)
    sh = np.fft.ifftshift(polar_radius(size, exp, ctr))
    sh[0, 0] = 1  # DC term

    fres = sh * fres
    fres = np.fft.ifft2(fres)

    if abs(fres.imag).max() > 1e-10:
        print('Symmetry error in creating fractal')
    else:
        res = np.real(fres)
        res = res / np.sqrt(var(res))

    return res


def blue_noise(size, fract_dim=1):
    '''make blue noise

    Make a matrix of specified size containing blue noise with power
    spectral density of the form: f^(5-2*`fract_dim`).  Image variance
    is normalized to 1.0.

    Note that the power spectrum here is the reciprocal of the pink
    noises's power spectrum

    Arguments
    ---------
    size : `int` or `tuple`
        if an int, we assume the image should be of dimensions `(size,
        size)`. if a tuple, must be a 2-tuple of ints specifying the
        dimensions
    fract_dim : `float`
        the fractal dimension of the blue noise

    Returns
    -------
    res : `np.array`
        the blue noise

    '''
    if not hasattr(size, '__iter__'):
        size = (size, size)

    res = np.random.randn(size[0], size[1])
    fres = np.fft.fft2(res)

    exp = 2.5-fract_dim
    ctr = np.ceil((res.shape + np.ones(2))/2.)
    sh = np.fft.ifftshift(polar_radius(size, exp, ctr))
    sh[0, 0] = 1  # DC term

    fres = sh * fres
    fres = np.fft.ifft2(fres)

    if abs(fres.imag).max() > 1e-10:
        print('Symmetry error in creating fractal')
    else:
        res = np.real(fres)
        res = res / np.sqrt(var(res))

    return res



import functools
from operator import mul


def convert_pyr_coeffs_to_pyr(pyr_coeffs):
    """this function takes a 'new pyramid' and returns the coefficients as a list

    this is to enable backwards compatibility

    Parameters
    ----------
    pyr_coeffs : `dict`
        The `pyr_coeffs` attribute of a `pyramid`.

    Returns
    -------
    coeffs : `list`
        list of `np.array`, which contains the pyramid coefficients in each band, in order from
        bottom of the pyramid to top (going through the orientations in order)
    highpass : `np.array` or None
        either the residual highpass from the pyramid or, if that doesn't exist, None
    lowpass : `np.array` or None
        either the residual lowpass from the pyramid or, if that doesn't exist, None

    """
    highpass = pyr_coeffs.pop('residual_highpass', None)
    lowpass = pyr_coeffs.pop('residual_lowpass', None)
    coeffs = [i[1] for i in sorted(pyr_coeffs.items(), key=lambda x: x[0])]
    return coeffs, highpass, lowpass


def max_pyr_height(imsz, filtsz):
    '''Compute maximum pyramid height for given image and filter sizes.

    Specifically, this computes the number of corrDn operations that can be sequentially performed
    when subsampling by a factor of 2.

    Parameters
    ----------
    imsz : `tuple` or `int`
        the size of the image (should be 2-tuple if image is 2d, `int` if it's 1d)
    filtsz : `tuple` or `int`
        the size of the filter (should be 2-tuple if image is 2d, `int` if it's 1d)

    Returns
    -------
    max_pyr_height : `int`
        The maximum height of the pyramid
    '''
    # check if inputs are one of int, tuple and have consistent type
    assert (isinstance(imsz, int) and isinstance(filtsz, int)) or (
            isinstance(imsz, tuple) and isinstance(filtsz, tuple))
    # 1D image case: reduce to the integer case
    if isinstance(imsz, tuple) and (len(imsz) == 1 or 1 in imsz):
        imsz = functools.reduce(mul, imsz)
        filtsz = functools.reduce(mul, filtsz)
    # integer case
    if isinstance(imsz, int):
        if imsz < filtsz:
            return 0
        else:
            return 1 + max_pyr_height(imsz // 2, filtsz)
    # 2D image case
    if isinstance(imsz, tuple):
        if min(imsz) < max(filtsz):
            return 0
        else:
            return 1 + max_pyr_height((imsz[0] // 2, imsz[1] // 2), filtsz)



import warnings
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.figure import Figure
from matplotlib import animation
from ..pyramids import convert_pyr_coeffs_to_pyr


class PyrFigure(Figure):
    """custom figure class to ensure that plots are created and saved with a constant dpi

    NOTE: generally, you shouldn't use this directly, relying instead on the make_figure
    function.

    If you do want to use, do the following: fig = plt.figure(FigureClass=PyrFigure) (NOT fig =
    PyrFigure()) and analogously for other pyplot functions (plt.subplots, etc)

    this enables us to make sure there's no aliasing: a single value in the (image) array that
    we're plotting will be represented as an integer multiple of pixels in the displayed figure

    The dpi that's chosen is an arbitrary value, the only thing that matters is that we use the
    same one when creating and saving the figure, which is what we ensure here. This also means
    that you will be unable to use plt.tight_layout, since we set the spacing and size of the
    subplots very intentionally.
    """
    def __init__(self, dpi=96, *args, **kwargs):
        kwargs['dpi'] = dpi
        Figure.__init__(self, *args, **kwargs)

    def savefig(self, fname, dpi_multiple=1, **kwargs):
        """Save the current figure.

        Call signature::

          savefig(fname, dpi_multiple=1, facecolor='w', edgecolor='w',
                  orientation='portrait', papertype=None, format=None,
                  transparent=False, bbox_inches=None, pad_inches=0.1,
                  frameon=None)

        The output formats available depend on the backend being used.

        Arguments
        ----------

        fname : str or file-like object
            A string containing a path to a filename, or a Python
            file-like object, or possibly some backend-dependent object
            such as :class:`~matplotlib.backends.backend_pdf.PdfPages`.

            If *format* is *None* and *fname* is a string, the output
            format is deduced from the extension of the filename. If
            the filename has no extension, the value of the rc parameter
            ``savefig.format`` is used.

            If *fname* is not a string, remember to specify *format* to
            ensure that the correct backend is used.

        Other Arguments
        ----------------

        dpi_multiple : [ scalar integer > 0 ]
            How to scale the figure's dots per inch (must be an integer to
            prevent aliasing). Default is 1, equivalent to using to the value
            of the figure (default matplotlib savefig behavior with dpi='figure')

        facecolor : color spec or None, optional
            the facecolor of the figure; if None, defaults to savefig.facecolor

        edgecolor : color spec or None, optional
            the edgecolor of the figure; if None, defaults to savefig.edgecolor

        orientation : {'landscape', 'portrait'}
            not supported on all backends; currently only on postscript output

        papertype : str
            One of 'letter', 'legal', 'executive', 'ledger', 'a0' through
            'a10', 'b0' through 'b10'. Only supported for postscript
            output.

        format : str
            One of the file extensions supported by the active
            backend.  Most backends support png, pdf, ps, eps and svg.

        transparent : bool
            If *True*, the axes patches will all be transparent; the
            figure patch will also be transparent unless facecolor
            and/or edgecolor are specified via kwargs.
            This is useful, for example, for displaying
            a plot on top of a colored background on a web page.  The
            transparency of these patches will be restored to their
            original values upon exit of this function.

        frameon : bool
            If *True*, the figure patch will be colored, if *False*, the
            figure background will be transparent.  If not provided, the
            rcParam 'savefig.frameon' will be used.

        bbox_inches : str or `~matplotlib.transforms.Bbox`, optional
            Bbox in inches. Only the given portion of the figure is
            saved. If 'tight', try to figure out the tight bbox of
            the figure. If None, use savefig.bbox

        pad_inches : scalar, optional
            Amount of padding around the figure when bbox_inches is
            'tight'. If None, use savefig.pad_inches

        bbox_extra_artists : list of `~matplotlib.artist.Artist`, optional
            A list of extra artists that will be considered when the
            tight bbox is calculated.
        """
        dpi = kwargs.pop('dpi', None)
        if dpi is not None:
            warnings.warn("Ignoring dpi argument: with PyrFigure, we do not use the dpi argument"
                          " for saving, use dpi_multiple instead (this is done to prevent "
                          "aliasing)")
        kwargs['dpi'] = self.dpi * dpi_multiple
        super().savefig(fname, **kwargs)

    def tight_layout(self, renderer=None, pad=1.08, h_pad=None, w_pad=None,
                     rect=None):
        """THIS IS NOT SUPPORTED (we control placement very specifically)
        """
        raise AttributeError("tight_layout is not supported with PyrFigure (we control the "
                             "layout, size, and spacing of these figures quite specifically and "
                             "don't want this kind of automatic changes)")


def make_figure(n_rows, n_cols, axis_size_pix, col_margin_pix=10, row_margin_pix=10, vert_pct=.8):
    """make a nice figure.

    this uses our custom PyrFigure class under the hood.

    Arguments
    ---------
    n_rows : `int`
        the number of rows to create in the figure
    n_cols : `int`
        the number of columns to create in the figure
    axis_size_pix : `tuple`
        2-tuple of ints specifying the size of each axis in the figure in pixels.
    col_margin_pix : `int`
        the number of pixels to leave between each column of axes
    row_margin_pix : `int`
        the number of pixels to leave between each row of axes
    vert_pct : `float`
        must lie between 0 and 1. if less than 1, we leave a little extra room at the top to allow
        a title. for example, if .8, then we add an extra 20% on top to leave room for a title

    Returns
    -------
    fig : `PyrFigure`
        the figure containing the axes at the specified size.
    """
    # this is an arbitrary value
    ppi = 96

    # we typically add extra space to the y direction to leave room for the title. this is
    # controlled by vert_pct: the default value works well if you want a title, and it should be 1
    # if you don't want to use a title
    fig = plt.figure(FigureClass=PyrFigure,
                     figsize=(((n_cols-1)*col_margin_pix+n_cols*axis_size_pix[1]) / ppi,
                              ((n_rows-1)*row_margin_pix+n_rows*(axis_size_pix[0]/vert_pct)) / ppi),
                     dpi=ppi)
    bbox = fig.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    fig_width, fig_height = bbox.width*fig.dpi, bbox.height*fig.dpi
    rel_axis_width = axis_size_pix[1] / fig_width
    rel_axis_height = axis_size_pix[0] / fig_height
    rel_col_margin = col_margin_pix / fig_width
    rel_row_margin = row_margin_pix / fig_height

    for i in range(n_rows):
        for j in range(n_cols):
            fig.add_axes([j*(rel_axis_width+rel_col_margin),
                          1.-((i+1)*rel_axis_height/vert_pct+i*rel_row_margin), rel_axis_width,
                          rel_axis_height], frameon=False, xticks=[], yticks=[])
    return fig


def _showIm(img, ax, vrange, zoom, title='', cmap=cm.gray, **kwargs):
    """helper function to display the image on the specified axis

    NOTE: should not be used directly.
    """
    ax.imshow(img, cmap=cmap, vmin=vrange[0], vmax=vrange[1], interpolation='none', **kwargs)

    if title is not None:
        # adapt the precision of displayed range to the order of magnitude of the values
        title = title + '\n range: [{:.1e}, {:.1e}] \n dims: [{}, {}] * {}'
        # 12 pt font looks good on axes that 256 pixels high, so we stick with that ratio
        ax.set_title(title.format(vrange[0], vrange[1], img.shape[0], img.shape[1], zoom),
                     {'fontsize': ax.bbox.height*(12./256)})


def reshape_axis(ax, axis_size_pix):
    """reshape axis to the specified size in pixels

    this will reshape an axis so that the given axis is the specified size in pixels, which we use
    to make sure that an axis is the same size as (or an integer multiple of) the array we're
    trying to display. this is to prevent aliasing

    NOTE: this can only shrink a big axis, not make a small one bigger, and will throw an exception
    if you try to do that.

    Arguments
    ---------
    ax : `matpotlib.pyplot.axis`
        the axis to reshape
    axis_size_pix : `int`
        the target size of the axis, in pixels

    Returns
    -------
    ax : `matplotlib.pyplot.axis`
        the reshaped axis
    """
    if ax.bbox.width < axis_size_pix[1] or ax.bbox.height < axis_size_pix[0]:
        raise Exception("Your axis is too small! Axis size: ({}, {}). Image size: ({}, {})".format(
            ax.bbox.width, ax.bbox.height, axis_size_pix[1], axis_size_pix[0]))
    bbox = ax.figure.get_window_extent().transformed(ax.figure.dpi_scale_trans.inverted())
    fig_width, fig_height = bbox.width*ax.figure.dpi, bbox.height*ax.figure.dpi
    rel_axis_width = axis_size_pix[1] / fig_width
    rel_axis_height = axis_size_pix[0] / fig_height
    ax.set_position([*ax.get_position().bounds[:2], rel_axis_width, rel_axis_height])
    return ax


def colormap_range(image, vrange='indep1', cmap=None):
    """Find the appropriate ranges for colormaps of provided images

    Arguments
    ---------
    image : `np.array` or `list`
        the image(s) to be shown. should be a 2d array (one image to display),
        3d array (multiple images to display, images are indexed along the first
        dimension), or list of 2d arrays. all images will be automatically
        rescaled so they're displayed at the same size. thus, their sizes must
        be scalar multiples of each other.
    vrange : `tuple` or `str`
        If a 2-tuple, specifies the image values vmin/vmax that are mapped to
        (ie. clipped to) the minimum and maximum value of the colormap,
        respectively.
        If a string:
        * `'auto0'`: all images have same vmin/vmax, which have the same
                     absolute value, and come from the minimum or maximum across
                     all images, whichever has the larger absolute value
        * `'auto1'`: all images have same vmin/vmax, which are the minimum/
                     maximum values across all images
        * `'auto2'`: all images have same vmin/vmax, which are the mean (across
                     all images) minus/plus 2 std dev (across all images)
        * `'auto3'`: all images have same vmin/vmax, chosen so as to map the
                     10th/90th percentile values to the 10th/90th percentile of
                     the display intensity range. For example: vmin is the 10th
                     percentile image value minus 1/8 times the difference
                     between the 90th and 10th percentile
        * `'indep0'`: each image has an independent vmin/vmax, which have the
                     same absolute value, which comes from either their minimum
                     or maximum value, whichever has the larger absolute value.
        * `'indep1'`: each image has an independent vmin/vmax, which are their
                     minimum/maximum values
        * `'indep2'`: each image has an independent vmin/vmax, which is their
                     mean minus/plus 2 std dev
        * `'indep3'`: each image has an independent vmin/vmax, chosen so that
                     the 10th/90th percentile values map to the 10th/90th
                     percentile intensities.

    Returns
    -------
    vrange_list : `list`
        list of tuples, same length as `image`. contains the (vmin, vmax) tuple
        for each image.

    """
    # flatimg is one long 1d array, which enables the min, max, mean, std, and percentile calls to
    # operate on the values from each of the images simultaneously.
    flatimg = np.concatenate([i.flatten() for i in image]).flatten()

    if isinstance(vrange, str):
        if vrange[:4] == 'auto':
            if vrange == 'auto0':
                M = np.nanmax([np.abs(np.nanmin(flatimg)), np.abs(np.nanmax(flatimg))])
                vrange_list = [-M, M]
            elif vrange == 'auto1' or vrange == 'auto':
                vrange_list = [np.nanmin(flatimg), np.nanmax(flatimg)]
            elif vrange == 'auto2':
                vrange_list = [np.nanmean(flatimg) - 2 * np.nanstd(flatimg),
                               np.nanmean(flatimg) + 2 * np.nanstd(flatimg)]
            elif vrange == 'auto3':
                p1 = np.nanpercentile(flatimg, 10)
                p2 = np.nanpercentile(flatimg, 90)
                vrange_list = [p1-(p2-p1)/8.0, p2+(p2-p1)/8.0]

            # make sure to return as many ranges as there are images
            vrange_list = [vrange_list] * len(image)

        elif vrange[:5] == 'indep':
            # independent vrange from recursive calls of this function per image
            vrange_list = [colormap_range(im, vrange.replace('indep', 'auto')
                           )[0][0] for im in image]
        else:
            vrange_list, _ = colormap_range(image, vrange='auto1')
            warnings.warn('Unknown vrange argument, using auto1 instead')
    else:
        # two numbers were passed, either as a list or tuple
        if len(vrange) != 2:
            raise Exception("If you're passing numbers to vrange,"
                            "there must be 2 of them!")
        vrange_list = [tuple(vrange)] * len(image)

    # double check that we're returning the right number of vranges
    assert len(image) == len(vrange_list)

    if cmap is None:
        if '0' in vrange:
            cmap = cm.RdBu_r
        else:
            cmap = cm.gray

    return vrange_list, cmap


def _check_shapes(images, video=False):
    """Helper function to check whether images can be zoomed in appropriately.

    this convenience function takes a list of images and finds out if they can all be displayed at
    the same size. for this to be the case, there must be an integer for each image such that the
    image can be multiplied by that integer to be the same size as the biggest image.

    Arguments
    ---------
    images : `list`
        list of numpy arrays to check the size of. In practice, these are 1d or 2d, but can in
        principle be any number of dimensions
    video: bool, optional (default False)
        handling signals in both space and time or only space.

    Returns
    -------
    max_shape : `tuple`
        2-tuple of integers, showing the shape of the largest image in the list

    Raises
    ------
    ValueError :
        If the images cannot be zoomed to the same. that is, if there is not an integer
        for each image such that the image can be multiplied by that integer to be the
        same size as the biggest image.
    """
    def check_shape_1d(shapes):
        max_shape = np.max(shapes)
        for s in shapes:
            if not (max_shape % s) == 0:
                raise ValueError("All images must be able to be 'zoomed in' to the largest image."
                                 "That is, the largest image must be a scalar multiple of all "
                                 "images.")
        return max_shape
    if video:
        time_dim = 1
    else:
        time_dim = 0
    max_shape = []
    for i in range(2):
        max_shape.append(check_shape_1d([img.shape[i+time_dim] for img in images]))
    return max_shape


def find_zooms(images, video=False):
    """find the zooms necessary to display a list of images

    Arguments
    ---------
    images : `list`
        list of numpy arrays to check the size of. In practice, these are 1d or 2d, but can in
        principle be any number of dimensions
    video: bool, optional (default False)
        handling signals in both space and time or only space.

    Returns
    -------
    zooms : `list`
        list of integers showing how much each image needs to be zoomed
    max_shape : `tuple`
        2-tuple of integers, showing the shape of the largest image in the list

    Raises
    ------
    ValueError :
        If the images cannot be zoomed to the same. that is, if there is not an integer
        for each image such that the image can be multiplied by that integer to be the
        same size as the biggest image.
    ValueError :
        If the two image dimensions require different levels of zoom (e.g., if the
        height must be zoomed by 2 but the width must be zoomed by 3).

    """
    max_shape = _check_shapes(images, video)
    if video:
        time_dim = 1
    else:
        time_dim = 0
    zooms = []
    for img in images:
        # this checks that there's only one unique value in the list
        # max_shape[i] // img.shape[i], where i indexes through the dimensions;
        # that is, that we zoom each dimension by the same amount. this should
        # then work with an arbitrary number of dimensions (in practice, 1 or
        # 2). by using max_shape instead of img.shape, we will only ever check
        # the first two non-time dimensions (so we'll ignore the RGBA channel
        # if any image has that)
        if len(set([s // img.shape[i+time_dim] for i, s in enumerate(max_shape)])) > 1:
            raise ValueError("Both height and width must be multiplied by same amount but got "
                             "image shape {} and max_shape {}!".format(img.shape, max_shape))
        zooms.append(max_shape[0] // img.shape[0])
    return zooms, max_shape


def _convert_signal_to_list(signal):
    """Convert signal to list.

    signal can be an array or a list of arrays. this guarantees it's a list,
    and raises an Exception if it's not an array or list.

    if it's an already a list, we don't check whether each value is a
    properly-shaped array, nor do we check the shape or dimensionality of a
    single array. these both happen in _process_signal

    Parameters
    ----------
    signal : np.ndarray or list
        the array or list of arrays to convert

    Returns
    -------
    signal : list

    """
    try:
        if isinstance(signal, np.ndarray):
            # then this is a single signal
            signal = [signal]
    except AttributeError:
        # then this is a list and we don't do anything
        pass
    if not isinstance(signal, list):
        raise TypeError("image must be a np.ndarray or a list! {} is unsupported".format(type(signal)))
    return signal


def _convert_title_to_list(title, signal):
    """Convert title to list and get vert_pct.

    This function makes sure that title is a list with the right number of
    elements, and sets vert_pct based on whether title is None or not

    Parameters
    ----------
    title : str, list, or None
        the title to check
    signal : list
        the signal (e.g., images) that will be plotted. must already have gone
        through _convert_signal_to_list

    Returns
    -------
    title : list
        list of title
    vert_pct : float
        how much of the axis should contain the image

    """
    if title is None:
        vert_pct = 1
    else:
        vert_pct = .8
    if not isinstance(title, list):
        title = len(signal) * [title]
    else:
        assert len(signal) == len(title), "Must have same number of titles and images!"
    return title, vert_pct


def _process_signal(signal, title, plot_complex, video=False):
    """Process signal and title for plotting.

    Two goals of this function:

    1. Check the shape of each image to make sure they look like either
       grayscale or RGB(A) images and raise an Exception if they don't

    2. Process complex images for proper plotting, splitting them up
       as specified by `plot_complex`

    Parameters
    ----------
    signal : list
        list of arrays to examine
    title : list
        list containing strs or Nones, for accompanying the images
    plot_complex : {'rectangular', 'polar', 'logpolar'}
        how to plot complex arrays
    video: bool, optional (default False)
        handling signals in both space and time or only space.

    Returns
    -------
    signal : list
        list of arrays containing the signal, ready to plot
    title : list
        list of titles, ready to plot
    contains_rgb : bool
        if at least one of the images is 3d (and thus RGB), this will be True.

    """
    if video:
        time_dim = 1
        sigtype = 'video'
    else:
        time_dim = 0
        sigtype = 'image'
    signal_tmp = []
    title_tmp = []
    contains_rgb = False
    for sig, t in zip(signal, title):
        if sig.ndim == (3 + time_dim):
            if sig.shape[-1] not in [3, 4]:
                raise Exception(
                    f"Can't figure out how to plot {sigtype} with shape {sig.shape} "
                    "as RGB(A)! RGB(A) signals should have their final"
                    "dimension of shape 3 or 4."
                )
            contains_rgb = True
        elif sig.ndim != (2 + time_dim):
            raise Exception(
                f"Can't figure out how to plot {sigtype} with "
                f"shape {sig.shape}! {sigtype.capitalize()}s should be be either "
                f"{2 + time_dim}d (grayscale) or {3 + time_dim}d (RGB(A), last "
                "dimension with 3 or 4 elements)."
            )
        if np.iscomplexobj(sig):
            if plot_complex == 'rectangular':
                signal_tmp.extend([np.real(sig), np.imag(sig)])
                if t is not None:
                    title_tmp.extend([t + " real", t + " imaginary"])
                else:
                    title_tmp.extend([None, None])
            elif plot_complex == 'polar':
                signal_tmp.extend([np.abs(sig), np.angle(sig)])
                if t is not None:
                    title_tmp.extend([t + " amplitude", t + " phase"])
                else:
                    title_tmp.extend([None, None])
            elif plot_complex == 'logpolar':
                signal_tmp.extend([np.log(1 + np.abs(sig)), np.angle(sig)])
                if t is not None:
                    title_tmp.extend([t + " log(1+amplitude)", t + " phase"])
                else:
                    title_tmp.extend([None, None])
        else:
            signal_tmp.append(np.asarray(sig))
            title_tmp.append(t)
    return signal_tmp, title_tmp, contains_rgb


def _check_zooms(signal, zoom, contains_rgb, video=False):
    """Check that all images can be zoomed correctly.

    Make sure that all images can be zoomed so they end up the same size, and
    figure out how to do that

    Parameters
    ----------
    signal : list
        list of arrays that will be plotted
    zoom : float
        how we're going to zoom the image
    contains_rgb : bool
        whether image contains at least one image to plot as RGB. This only
        matters when we're given a 3d array and we want to know whether it was
        supposed to be a single RGB image or multiple grayscale ones
    video: bool, optional (default False)
        handling signals in both space and time or just space.

    Returns
    -------
    zooms : np.ndarray
        how much to zoom each image
    max_shape : np.ndarray
        contains 2 ints, giving the max image size in pixels

    """
    if video:
        time_dim = 1
    else:
        time_dim = 0

    if hasattr(zoom, '__iter__'):
        raise Exception("zoom must be a single number!")
    if all([sig.shape == signal[0].shape for sig in signal]):
        # then this is one or multiple images/videos, all the same size
        zooms = [1 for i in signal]
        if contains_rgb:
            max_shape = signal[0].shape[-3:-1]
        else:
            max_shape = signal[0].shape[-2:]
    else:
        # then we have multiple images/videos that are different shapes
        zooms, max_shape = find_zooms(signal, video)
    max_shape = np.asarray(max_shape)
    zooms = zoom * np.asarray(zooms)
    if not ((zoom * max_shape).astype(int) == zoom * max_shape).all():
        raise Exception("zoom * signal.shape must result in integers!")
    return zooms, max_shape


def _setup_figure(ax, col_wrap, image, zoom, max_shape, vert_pct):
    """Create figure with appropriate arrangement and size of axes

    Creates (or tries to resize) set of axes for the appropriate arguments

    """
    if ax is None:
        if col_wrap is None:
            n_cols = len(image)
            n_rows = 1
        else:
            n_cols = col_wrap
            n_rows = int(np.ceil(len(image) / n_cols))
        fig = make_figure(n_rows, n_cols, zoom * max_shape, vert_pct=vert_pct)
        axes = fig.axes
    else:
        fig = ax.figure
        axes = [reshape_axis(ax,  zoom * max_shape)]
    return fig, axes


def imshow(image, vrange='indep1', zoom=1, title='', col_wrap=None, ax=None,
           cmap=None, plot_complex='rectangular', **kwargs):
    """Show image(s).

    Arguments
    ---------
    image : `np.array` or `list`
        the image(s) to plot. Images can be either grayscale, in which case
        they must be 2d arrays of shape `(h,w)`, or RGB(A), in which case they
        must be 3d arrays of shape `(h,w,c)` where `c` is 3 (for RGB) or 4 (to
        also plot the alpha channel). If multiple images, must be a list of
        such arrays (note this means we do not support an array of shape
        `(n,h,w)` for multiple grayscale images). all images will be
        automatically rescaled so they're displayed at the same size. thus,
        their sizes must be scalar multiples of each other.
    vrange : `tuple` or `str`
        If a 2-tuple, specifies the image values vmin/vmax that are mapped to
        the minimum and maximum value of the colormap, respectively. If a
        string:

        * `'auto0'`: all images have same vmin/vmax, which have the same absolute
                     value, and come from the minimum or maximum across all
                     images, whichever has the larger absolute value
        * `'auto/auto1'`: all images have same vmin/vmax, which are the
                          minimum/maximum values across all images
        * `'auto2'`: all images have same vmin/vmax, which are the mean (across
                     all images) minus/ plus 2 std dev (across all images)
        * `'auto3'`: all images have same vmin/vmax, chosen so as to map the
                     10th/90th percentile values to the 10th/90th percentile of
                     the display intensity range. For example: vmin is the 10th
                     percentile image value minus 1/8 times the difference
                     between the 90th and 10th percentile
        * `'indep0'`: each image has an independent vmin/vmax, which have the
                      same absolute value, which comes from either their
                      minimum or maximum value, whichever has the larger
                      absolute value.
        * `'indep1'`: each image has an independent vmin/vmax, which are their
                      minimum/maximum values
        * `'indep2'`: each image has an independent vmin/vmax, which is their
                      mean minus/plus 2 std dev
        * `'indep3'`: each image has an independent vmin/vmax, chosen so that
                      the 10th/90th percentile values map to the 10th/90th
                      percentile intensities.
    zoom : `float`
        ratio of display pixels to image pixels. if >1, must be an integer. If
        <1, must be 1/d where d is a a divisor of the size of the largest
        image.
    title : `str`, `list`, or None, optional
        Title for the plot. In addition to the specified title, we add a
        subtitle giving the plotted range and dimensionality (with zoom)

        * if `str`, will put the same title on every plot.
        * if `list`, all values must be `str`, must be the same length as img,
          assigning each title to corresponding image.
        * if `None`, no title will be printed (and subtitle will be removed;
          unsupported for complex tensors).
    col_wrap : `int` or None, optional
        number of axes to have in each row. If None, will fit all axes in a
        single row.
    ax : `matplotlib.pyplot.axis` or None, optional
        if None, we make the appropriate figure. otherwise, we resize the axes
        so that it's the appropriate number of pixels (done by shrinking the
        bbox - if the bbox is already too small, this will throw an Exception!,
        so first define a large enough figure using either make_figure or
        plt.figure)
    cmap : matplotlib colormap, optional
        colormap to use when showing these images
    plot_complex : {'rectangular', 'polar', 'logpolar'}, optional
        specifies handling of complex values.

        * `'rectangular'`: plot real and imaginary components as separate images
        * `'polar'`: plot amplitude and phase as separate images
        * `'logpolar'`: plot `ln(1+ amplitude)` and phase as separate images.
        The compressive non-linear contrast function applied to amplitude is
        intended as a visualization step to avoid the large intensity
        components from dominating.
    kwargs :
        Passed to `ax.imshow`

    Returns
    -------
    fig : `PyrFigure`
        figure containing the plotted images

    """
    if plot_complex not in ['rectangular', 'polar', 'logpolar']:
        raise Exception("Don't know how to handle plot_complex value "
                        "{}!".format(plot_complex))

    # Make sure image is a list, do some preliminary checks
    image = _convert_signal_to_list(image)

    # want to do this check before converting title to a list (at which
    # point `title is None` will always be False). we do it here instad
    # of checking whether the first item of title is None because it's
    # conceivable that the user passed `title=[None, 'important
    # title']`, and in that case we do want the space for the title
    title, vert_pct = _convert_title_to_list(title, image)

    # Process complex images for plotting, double-check image size to see if we
    # have RGB(A) images
    image, title, contains_rgb = _process_signal(image, title, plot_complex)

    # make sure we can properly zoom all images
    zooms, max_shape = _check_zooms(image, zoom, contains_rgb)

    # get the figure and axes created
    fig, axes = _setup_figure(ax, col_wrap, image, zoom, max_shape, vert_pct)

    vrange_list, cmap = colormap_range(image=image, vrange=vrange, cmap=cmap)

    for im, a, r, t, z in zip(image, axes, vrange_list, title, zooms):
        _showIm(im, a, r, z, t, cmap, **kwargs)

    return fig


def animshow(video, framerate=2., as_html5=True, repeat=False,
             vrange='indep1', zoom=1, title='', col_wrap=None, ax=None,
             cmap=None, plot_complex='rectangular', **kwargs):
    """Display one or more videos (3d array) as a matplotlib animation or an HTML video.

    Arguments
    ---------
    video : `np.array` or `list`
        the videos(s) to show. Videos can be either grayscale, in which case
        they must be 3d arrays of shape `(f,h,w)`, or RGB(A), in which case
        they must be 4d arrays of shape `(f,h,w,c)` where `c` is 3 (for RGB) or
        4 (to also plot the alpha channel) and `f` indexes frames. If multiple
        videos, must be a list of such arrays (note this means we do not
        support an array of shape `(n,f,h,w)` for multiple grayscale videos).
        all videos will be automatically rescaled so they're displayed at the
        same size. thus, their sizes must be scalar multiples of each other. If
        multiple videos, all must have the same number of frames (first
        dimension).
    framerate : `float`
        Temporal resolution of the video, in Hz (frames per second).
    as_html : `bool`
        If True, return an HTML5 video; otherwise return the underying matplotlib animation object
        (e.g. to save to .gif). should set to True to display in a Jupyter notebook.
        Requires ipython to be installed.
    repeat : `bool`
        whether to loop the animation or just play it once
    vrange : `tuple` or `str`
        If a 2-tuple, specifies the image values vmin/vmax that are mapped to the minimum and
        maximum value of the colormap, respectively. If a string:

        * `'auto/auto1'`: all images have same vmin/vmax, which are the minimum/maximum values
                          across all images
        * `'auto2'`: all images have same vmin/vmax, which are the mean (across all images) minus/
                     plus 2 std dev (across all images)
        * `'auto3'`: all images have same vmin/vmax, chosen so as to map the 10th/90th percentile
                     values to the 10th/90th percentile of the display intensity range. For
                     example: vmin is the 10th percentile image value minus 1/8 times the
                     difference between the 90th and 10th percentile
        * `'indep1'`: each image has an independent vmin/vmax, which are their minimum/maximum
                      values
        * `'indep2'`: each image has an independent vmin/vmax, which is their mean minus/plus 2
                      std dev
        * `'indep3'`: each image has an independent vmin/vmax, chosen so that the 10th/90th
                      percentile values map to the 10th/90th percentile intensities.
    zoom : `float`
        amount we zoom the video frames (must result in an integer when multiplied by
        video.shape[1:])
    title : `str` , `list` or None
        Title for the plot:

        * if `str`, will put the same title on every plot.
        * if `list`, all values must be `str`, must be the same length as img, assigning each
          title to corresponding image.
        * if None, no title will be printed.
    col_wrap : `int` or None, optional
        number of axes to have in each row. If None, will fit all axes in a
        single row.
    ax : `matplotlib.pyplot.axis` or None, optional
        if None, we make the appropriate figure. otherwise, we resize the axes
        so that it's the appropriate number of pixels (done by shrinking the
        bbox - if the bbox is already too small, this will throw an Exception!,
        so first define a large enough figure using either make_figure or
        plt.figure)
    cmap : matplotlib colormap, optional
        colormap to use when showing these images
    plot_complex : {'rectangular', 'polar', 'logpolar'}, optional
        specifies handling of complex values.

        * `'rectangular'`: plot real and imaginary components as separate images
        * `'polar'`: plot amplitude and phase as separate images
        * `'logpolar'`: plot log_2 amplitude and phase as separate images
    kwargs :
        Passed to `ax.imshow`

    Returns
    -------
    anim : HTML object or FuncAnimation object
        Animation, format depends on `as_html`.

    """
    if as_html5:
        try:
            from IPython.display import HTML
        except ImportError:
            raise ImportError("Unable to import IPython.display.HTML, animshow must be called with "
                              "as_html5=False")
    video = _convert_signal_to_list(video)
    video_n_frames = np.asarray([v.shape[0] for v in video])
    if (video_n_frames != video_n_frames[0]).any():
        raise Exception("All videos must have the same number of frames! But you "
                        "passed videos with {} frames".format(video_n_frames))
    title, vert_pct = _convert_title_to_list(title, video)
    video, title, contains_rgb = _process_signal(video, title, plot_complex, video=True)
    zooms, max_shape = _check_zooms(video, zoom, contains_rgb, video=True)
    fig, axes = _setup_figure(ax, col_wrap, video, zoom, max_shape, vert_pct)
    vrange_list, cmap = colormap_range(image=video, vrange=vrange, cmap=cmap)

    first_image = [v[0] for v in video]
    for im, a, r, t, z in zip(first_image, axes, vrange_list, title, zooms):
        _showIm(im, a, r, z, t, cmap, **kwargs)

    artists = [fig.axes[i].images[0] for i in range(len(fig.axes))]

    for i, a in enumerate(artists):
        a.set_clim(vrange_list[i])

    def animate_video(t):
        for i, a in enumerate(artists):
            frame = video[i][t].astype(float)
            a.set_data(frame)
        return artists

    # Produce the animation
    anim = animation.FuncAnimation(fig, frames=len(video[0]),
                                   interval=1000/framerate, blit=True,
                                   func=animate_video, repeat=repeat,
                                   repeat_delay=500)

    plt.close(fig)

    if as_html5:
        # to_html5_video will call savefig with a dpi kwarg, so our custom figure class will raise
        # a warning. we don't want to worry people, so we go ahead and suppress it
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            return HTML(anim.to_html5_video())
    return anim


def pyrshow(pyr_coeffs, is_complex=False, vrange='indep1', col_wrap=None, zoom=1, show_residuals=True, **kwargs):
    """Display the coefficients of the pyramid in an orderly fashion

    Arguments
    ---------
    pyr_coeffs : `dict`
        from the pyramid object (i.e. pyr.pyr_coeffs)
    is_complex : `bool`
        default False, indicates whether the pyramids is real or complex
        indicating whether the pyramid is complex or real
    vrange : `tuple` or `str`
        If a 2-tuple, specifies the image values vmin/vmax that are mapped to the minimum and
        maximum value of the colormap, respectively. If a string:

        * `'auto/auto1'`: all images have same vmin/vmax, which are the minimum/maximum values
                          across all images
        * `'auto2'`: all images have same vmin/vmax, which are the mean (across all images) minus/
                     plus 2 std dev (across all images)
        * `'auto3'`: all images have same vmin/vmax, chosen so as to map the 10th/90th percentile
                     values to the 10th/90th percentile of the display intensity range. For
                     example: vmin is the 10th percentile image value minus 1/8 times the
                     difference between the 90th and 10th percentile
        * `'indep1'`: each image has an independent vmin/vmax, which are their minimum/maximum
                      values
        * `'indep2'`: each image has an independent vmin/vmax, which is their mean minus/plus 2
                      std dev
        * `'indep3'`: each image has an independent vmin/vmax, chosen so that the 10th/90th
                      percentile values map to the 10th/90th percentile intensities.
    col_wrap : `int` or None
        Only usable when the pyramid is one-dimensional (e.g., Gaussian or Laplacian Pyramid),
        otherwise the column wrap is determined by the number of bands. If not None, how many axes
        to have in a given row.
    zoom : `float`
        how much to scale the size of the images by. zoom times the size of the largest image must
        be an integer (and thus zoom should probably be an integer or 1/(2^n)).
    show_residuals : `bool`
        whether to display the residual bands (lowpass, highpass depending on the pyramid type)

    any additional kwargs will be passed through to imshow.

    Returns
    -------
    fig: `PyrFigure`
        the figure displaying the coefficients.
    """
    # right now, we do *not* do this the same as the old code. Instead of taking the coefficients
    # and arranging them in a spiral, we use imshow and arrange them neatly, displaying all at the
    # same size (and zoom / original image size clear), with different options for vrange. It
    # doesn't seem worth it to me to implement a version that looks like the old one, since that
    # would be a fair amount of work for dubious payoff, but if you want to implement that, the way
    # to do it is to probably use gridspec (https://matplotlib.org/users/gridspec.html), arranging
    # the axes in an orderly way and then passing them through to imshow somehow, rather than
    # pasting all coefficients into a giant array.
    # and the steerable pyramids have a num_orientations attribute

    num_scales = np.max(np.asarray([k for k in pyr_coeffs.keys() if isinstance(k, tuple)])[:,0]) + 1
    num_orientations = np.max(np.asarray([k for k in pyr_coeffs.keys() if isinstance(k, tuple)])[:,1]) + 1

    col_wrap_new = num_orientations
    if is_complex:
        col_wrap_new *= 2
    # not sure about scope here, so we make sure to copy the
    # pyr_coeffs dictionary.
    imgs, highpass, lowpass = convert_pyr_coeffs_to_pyr(pyr_coeffs.copy())
    # we can similarly grab the labels for height and band
    # from the keys in this pyramid coefficients dictionary
    pyr_coeffs_keys = [k for k in pyr_coeffs.keys() if isinstance(k, tuple)]
    titles = ["height %02d, band %02d" % (h, b) for h, b in sorted(pyr_coeffs_keys)]
    if show_residuals:
        if highpass is not None:
            titles += ["residual highpass"]
            imgs.append(highpass)
        if lowpass is not None:
            titles += ["residual lowpass"]
            imgs.append(lowpass)
    if col_wrap_new is not None and col_wrap_new != 1:
        if col_wrap is None:
            col_wrap = col_wrap_new
    # if these are really 1d (i.e., have shape (1, x) or (x, 1)), then we want them to be 1d
    imgs = [i.squeeze() for i in imgs]
    if imgs[0].ndim == 1:
        # then we just want to plot each of the bands in a different subplot, no need to be fancy.
        if col_wrap is not None:
            warnings.warn("When the pyramid is 1d, we ignore col_wrap and just use "
                          "pyr.num_orientations to determine the number of columns!")
        height = num_scales
        if "residual highpass" in titles:
            height += 1
        if "residual lowpass" in titles:
            height += 1
        fig, axes = plt.subplots(height, num_orientations,
                                 figsize=(5*zoom, 5*zoom*num_orientations), **kwargs)
        plt.subplots_adjust(hspace=1.2, wspace=1.2)
        for i, ax in enumerate(axes.flatten()):
            ax.plot(imgs[i])
            ax.set_title(titles[i])
        return fig
    else:
        try:
            _check_shapes(imgs)
        except ValueError:
            err_scales = num_scales
            residual_err_msg = ""
            shapes = [(imgs[0].shape[0]/ 2**i, imgs[0].shape[1] / 2**i) for i in range(err_scales)]
            err_msg = [f"scale {i} shape: {sh}" for i, sh in enumerate(shapes)]
            if show_residuals:
                err_scales += 1
                residual_err_msg = ", plus 1 (for the residual lowpass)"
                shape = (imgs[0].shape[0]/ int(2**err_scales), imgs[0].shape[1] / int(2**err_scales))
                err_msg.append(f"residual lowpass shape: {shape}")
            err_msg = "\n".join(err_msg)
            raise ValueError("In order to correctly display pyramid coefficients, the shape of"
                             f" the initial image must be evenly divisible by two {err_scales} "
                             "times, where this number is the height of the "
                             f"pyramid{residual_err_msg}. "
                             f"Instead, found:\n{err_msg}")
        return imshow(imgs, vrange=vrange, col_wrap=col_wrap, zoom=zoom, title=titles, **kwargs)



import numpy as np
from .pyramid import SteerablePyramidBase
from .filters import parse_filter
from .c.wrapper import corrDn, upConv


class SteerablePyramidSpace(SteerablePyramidBase):
    """Steerable pyramid (using spatial convolutions)

    Notes
    -----
    Transform described in [1]_, filter kernel design described in [2]_.

    Parameters
    ----------
    image : `array_like`
        2d image upon which to construct to the pyramid.
    height : 'auto' or `int`.
        The height of the pyramid. If 'auto', will automatically determine based on the size of
        `image`.
    order : {0, 1, 3, 5}.
        The Gaussian derivative order used for the steerable filters. If you want a different
        value, see SteerablePyramidFreq. Note that to achieve steerability the minimum number
        of orientation is `order` + 1, and is used here. To get more orientations at the same
        order, use the method `steer_coeffs`
    edge_type : {'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend', 'dont-compute'}
        Specifies how to handle edges. Options are:

        * `'circular'` - circular convolution
        * `'reflect1'` - reflect about the edge pixels
        * `'reflect2'` - reflect, doubling the edge pixels
        * `'repeat'` - repeat the edge pixels
        * `'zero'` - assume values of zero outside image boundary
        * `'extend'` - reflect and invert
        * `'dont-compute'` - zero output when filter overhangs imput boundaries.

    Attributes
    ----------
    image : `array_like`
        The input image used to construct the pyramid.
    image_size : `tuple`
        The size of the input image.
    pyr_type : `str` or `None`
        Human-readable string specifying the type of pyramid. For base class, is None.
    edge_type : `str`
        Specifies how edges were handled.
    pyr_coeffs : `dict`
        Dictionary containing the coefficients of the pyramid. Keys are `(level, band)` tuples and
        values are 1d or 2d numpy arrays (same number of dimensions as the input image)
    pyr_size : `dict`
        Dictionary containing the sizes of the pyramid coefficients. Keys are `(level, band)`
        tuples and values are tuples.
    is_complex : `bool`
        Whether the coefficients are complex- or real-valued. Only `SteerablePyramidFreq` can have
        a value of True, all others must be False.

    References
    ----------
    .. [1] E P Simoncelli and W T Freeman, "The Steerable Pyramid: A Flexible Architecture for
       Multi-Scale Derivative Computation," Second Int'l Conf on Image Processing, Washington, DC,
       Oct 1995.
    .. [2] A Karasaridis and E P Simoncelli, "A Filter Design Technique for Steerable Pyramid
       Image Transforms", ICASSP, Atlanta, GA, May 1996.
    """

    def __init__(self, image, height='auto', order=1, edge_type='reflect1'):
        super().__init__(image=image, edge_type=edge_type)

        self.order = order
        self.num_orientations = self.order + 1
        self.filters = parse_filter("sp{:d}_filters".format(self.num_orientations-1), normalize=False)
        self.pyr_type = 'SteerableSpace'
        self._set_num_scales('lofilt', height)

        hi0 = corrDn(image=self.image, filt=self.filters['hi0filt'], edge_type=self.edge_type)

        self.pyr_coeffs['residual_highpass'] = hi0
        self.pyr_size['residual_highpass'] = hi0.shape

        lo = corrDn(image=self.image, filt=self.filters['lo0filt'], edge_type=self.edge_type)
        for i in range(self.num_scales):
            # assume square filters  -- start of buildSpyrLevs
            bfiltsz = int(np.floor(np.sqrt(self.filters['bfilts'].shape[0])))

            for b in range(self.num_orientations):
                filt = self.filters['bfilts'][:, b].reshape(bfiltsz, bfiltsz).T
                band = corrDn(image=lo, filt=filt, edge_type=self.edge_type)
                self.pyr_coeffs[(i, b)] = np.asarray(band)
                self.pyr_size[(i, b)] = band.shape

            lo = corrDn(image=lo, filt=self.filters['lofilt'], edge_type=self.edge_type, step=(2, 2))

        self.pyr_coeffs['residual_lowpass'] = lo
        self.pyr_size['residual_lowpass'] = lo.shape

    def recon_pyr(self, order=None, edge_type=None, levels='all', bands='all'):
        """Reconstruct the image, optionally using subset of pyramid coefficients.

        Parameters
        ----------
        order : {None, 0, 1, 3, 5}.
            the Gaussian derivative order you want to use for the steerable pyramid filters used to
            reconstruct the pyramid. If None, uses the same order as that used to construct the
            pyramid.
        edge_type : {None, 'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend',
                     'dont-compute'}
            Specifies how to handle edges. Options are:

            * None (default) - use `self.edge_type`, the edge_type used to construct the pyramid
            * `'circular'` - circular convolution
            * `'reflect1'` - reflect about the edge pixels
            * `'reflect2'` - reflect, doubling the edge pixels
            * `'repeat'` - repeat the edge pixels
            * `'zero'` - assume values of zero outside image boundary
            * `'extend'` - reflect and inverts
            * `'dont-compute'` - zero output when filter overhangs imput boundaries.
        levels : `list`, `int`,  or {`'all'`, `'residual_highpass'`}
            If `list` should contain some subset of integers from `0` to `self.num_scales-1`
            (inclusive) and `'residual_lowpass'`. If `'all'`, returned value will contain all
            valid levels. Otherwise, must be one of the valid levels.
        bands : `list`, `int`, or `'all'`.
            If list, should contain some subset of integers from `0` to `self.num_orientations-1`.
            If `'all'`, returned value will contain all valid orientations. Otherwise, must be one
            of the valid orientations.

        Returns
        -------
        recon : `np.array`
            The reconstructed image.
        """

        if order is None:
            filters = self.filters
            recon_keys = self._recon_keys(levels, bands)
        else:
            filters = parse_filter("sp{:d}_filters".format(order), normalize=False)
            recon_keys = self._recon_keys(levels, bands, order+1)

        # assume square filters  -- start of buildSpyrLevs
        bfiltsz = int(np.floor(np.sqrt(filters['bfilts'].shape[0])))

        if edge_type is None:
            edges = self.edge_type
        else:
            edges = edge_type


        # initialize reconstruction
        if 'residual_lowpass' in recon_keys:
            recon = self.pyr_coeffs['residual_lowpass']
        else:
            recon = np.zeros_like(self.pyr_coeffs['residual_lowpass'])

        for lev in reversed(range(self.num_scales)):
            # we need to upConv once per level, in order to up-sample
            # the image back to the right shape.
            recon = upConv(image=recon, filt=filters['lofilt'], edge_type=edges,
                           step=(2, 2), start=(0, 0), stop=self.pyr_size[(lev, 0)])
            # I think the most effective way to do this is to just
            # check every possible sub-band and then only add in the
            # ones we want (given that we have to loop through the
            # levels above in order to up-sample)
            for band in reversed(range(self.num_orientations)):
                if (lev, band) in recon_keys:
                    filt = filters['bfilts'][:, band].reshape(bfiltsz, bfiltsz, order='F')
                    recon += upConv(image=self.pyr_coeffs[(lev, band)], filt=filt, edge_type=edges,
                                    stop=self.pyr_size[(lev, band)])

        # apply lo0filt
        recon = upConv(image=recon, filt=filters['lo0filt'], edge_type=edges, stop=recon.shape)

        if 'residual_highpass' in recon_keys:
            recon += upConv(image=self.pyr_coeffs['residual_highpass'], filt=filters['hi0filt'],
                            edge_type=edges, start=(0, 0), step=(1, 1), stop=recon.shape)

        return recon



from .GaussianPyramid import GaussianPyramid
from .LaplacianPyramid import LaplacianPyramid
from .WaveletPyramid import WaveletPyramid
from .SteerablePyramidSpace import SteerablePyramidSpace
from .SteerablePyramidFreq import SteerablePyramidFreq
from .steer import steer, steer_to_harmonics_mtx
from .pyr_utils import convert_pyr_coeffs_to_pyr, max_pyr_height



import numpy as np
import warnings
from .pyr_utils import max_pyr_height
from .filters import named_filter
from .steer import steer


class Pyramid:
    """Base class for multiscale pyramids

    You should not instantiate this base class, it is instead inherited by the other classes found
    in this module.

    Parameters
    ----------
    image : `array_like`
        1d or 2d image upon which to construct to the pyramid.
    edge_type : {'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend', 'dont-compute'}
        Specifies how to handle edges. Options are:

        * `'circular'` - circular convolution
        * `'reflect1'` - reflect about the edge pixels
        * `'reflect2'` - reflect, doubling the edge pixels
        * `'repeat'` - repeat the edge pixels
        * `'zero'` - assume values of zero outside image boundary
        * `'extend'` - reflect and invert
        * `'dont-compute'` - zero output when filter overhangs imput boundaries.

    Attributes
    ----------
    image : `array_like`
        The input image used to construct the pyramid.
    image_size : `tuple`
        The size of the input image.
    pyr_type : `str` or `None`
        Human-readable string specifying the type of pyramid. For base class, is None.
    edge_type : `str`
        Specifies how edges were handled.
    pyr_coeffs : `dict`
        Dictionary containing the coefficients of the pyramid. Keys are `(level, band)` tuples and
        values are 1d or 2d numpy arrays (same number of dimensions as the input image)
    pyr_size : `dict`
        Dictionary containing the sizes of the pyramid coefficients. Keys are `(level, band)`
        tuples and values are tuples.
    is_complex : `bool`
        Whether the coefficients are complex- or real-valued. Only `SteerablePyramidFreq` can have
        a value of True, all others must be False.
    """

    def __init__(self, image, edge_type):

        self.image = np.asarray(image).astype(float)
        if self.image.ndim == 1:
            self.image = self.image.reshape(-1, 1)
        assert self.image.ndim == 2, "Error: Input signal must be 1D or 2D."

        self.image_size = self.image.shape
        if not hasattr(self, 'pyr_type'):
            self.pyr_type = None
        self.edge_type = edge_type
        self.pyr_coeffs = {}
        self.pyr_size = {}
        self.is_complex = False


    def _set_num_scales(self, filter_name, height, extra_height=0):
        """Figure out the number of scales (height) of the pyramid

        The user should not call this directly. This is called during construction of a pyramid,
        and is based on the size of the filters (thus, should be called after instantiating the
        filters) and the input image, as well as the `extra_height` parameter (which corresponds to
        the residuals, which the Gaussian pyramid contains and others do not).

        This sets `self.num_scales` directly instead of returning something, so be careful.

        Parameters
        ----------
        filter_name : `str`
            Name of the filter in the `filters` dict that determines the height of the pyramid
        height : `'auto'` or `int`
            During construction, user can specify the number of scales (height) of the pyramid.
            The pyramid will have this number of scales unless that's greater than the maximum
            possible height.
        extra_height : `int`, optional
            The automatically calculated maximum number of scales is based on the size of the input
            image and filter size. The Gaussian pyramid also contains the final residuals and so we
            need to add one more to this number.

        Returns
        -------
        None
        """
        # the Gaussian and Laplacian pyramids can go one higher than the value returned here, so we
        # use the extra_height argument to allow for that
        max_ht = max_pyr_height(self.image.shape, self.filters[filter_name].shape) + extra_height
        if height == 'auto':
            self.num_scales = max_ht
        elif height > max_ht:
            raise ValueError("Cannot build pyramid higher than %d levels." % (max_ht))
        else:
            self.num_scales = int(height)

    def _recon_levels_check(self, levels):
        """Check whether levels arg is valid for reconstruction and return valid version

        When reconstructing the input image (i.e., when calling `recon_pyr()`), the user specifies
        which levels to include. This makes sure those levels are valid and gets them in the form
        we expect for the rest of the reconstruction. If the user passes `'all'`, this constructs
        the appropriate list (based on the values of `self.pyr_coeffs`).

        Parameters
        ----------
        levels : `list`, `int`,  or {`'all'`, `'residual_highpass'`, or `'residual_lowpass'`}
            If `list` should contain some subset of integers from `0` to `self.num_scales-1`
            (inclusive) and `'residual_highpass'` and `'residual_lowpass'` (if appropriate for the
            pyramid). If `'all'`, returned value will contain all valid levels. Otherwise, must be
            one of the valid levels.

        Returns
        -------
        levels : `list`
            List containing the valid levels for reconstruction.

        """
        if isinstance(levels, str) and levels == 'all':
            levels = ['residual_highpass'] + list(range(self.num_scales)) + ['residual_lowpass']
        else:
            if not hasattr(levels, '__iter__') or isinstance(levels, str):
                # then it's a single int or string
                levels = [levels]
            levs_nums = np.asarray([int(i) for i in levels if isinstance(i, int) or i.isdigit()])
            assert (levs_nums >= 0).all(), "Level numbers must be non-negative."
            assert (levs_nums < self.num_scales).all(), "Level numbers must be in the range [0, %d]" % (self.num_scales-1)
            levs_tmp = list(np.sort(levs_nums))  # we want smallest first
            if 'residual_highpass' in levels:
                levs_tmp = ['residual_highpass'] + levs_tmp
            if 'residual_lowpass' in levels:
                levs_tmp = levs_tmp + ['residual_lowpass']
            levels = levs_tmp
        # not all pyramids have residual highpass / lowpass, but it's easier to construct the list
        # including them, then remove them if necessary.
        if 'residual_lowpass' not in self.pyr_coeffs.keys() and 'residual_lowpass' in levels:
            levels.pop(-1)
        if 'residual_highpass' not in self.pyr_coeffs.keys() and 'residual_highpass' in levels:
            levels.pop(0)
        return levels

    def _recon_bands_check(self, bands):
        """Check whether bands arg is valid for reconstruction and return valid version

        When reconstructing the input image (i.e., when calling `recon_pyr()`), the user specifies
        which orientations to include. This makes sure those orientations are valid and gets them
        in the form we expect for the rest of the reconstruction. If the user passes `'all'`, this
        constructs the appropriate list (based on the values of `self.pyr_coeffs`).

        Parameters
        ----------
        bands : `list`, `int`, or `'all'`.
            If list, should contain some subset of integers from `0` to `self.num_orientations-1`.
            If `'all'`, returned value will contain all valid orientations. Otherwise, must be one
            of the valid orientations.

        Returns
        -------
        bands: `list`
            List containing the valid orientations for reconstruction.
        """
        if isinstance(bands, str) and bands == "all":
            bands = np.arange(self.num_orientations)
        else:
            bands = np.array(bands, ndmin=1)
            assert (bands >= 0).all(), "Error: band numbers must be larger than 0."
            assert (bands < self.num_orientations).all(), "Error: band numbers must be in the range [0, %d]" % (self.num_orientations - 1)
        return bands

    def _recon_keys(self, levels, bands, max_orientations=None):
        """Make a list of all the relevant keys from `pyr_coeffs` to use in pyramid reconstruction

        When reconstructing the input image (i.e., when calling `recon_pyr()`), the user specifies
        some subset of the pyramid coefficients to include in the reconstruction. This function
        takes in those specifications, checks that they're valid, and returns a list of tuples
        that are keys into the `pyr_coeffs` dictionary.

        Parameters
        ----------
        levels : `list`, `int`,  or {`'all'`, `'residual_highpass'`, `'residual_lowpass'`}
            If `list` should contain some subset of integers from `0` to `self.num_scales-1`
            (inclusive) and `'residual_highpass'` and `'residual_lowpass'` (if appropriate for the
            pyramid). If `'all'`, returned value will contain all valid levels. Otherwise, must be
            one of the valid levels.
        bands : `list`, `int`, or `'all'`.
            If list, should contain some subset of integers from `0` to `self.num_orientations-1`.
            If `'all'`, returned value will contain all valid orientations. Otherwise, must be one
            of the valid orientations.
        max_orientations: `None` or `int`.
            The maximum number of orientations we allow in the reconstruction. when we determine
            which ints are allowed for bands, we ignore all those greater than max_orientations.

        Returns
        -------
        recon_keys : `list`
            List of `tuples`, all of which are keys in `pyr_coeffs`. These are the coefficients to
            include in the reconstruction of the image.

        """
        levels = self._recon_levels_check(levels)
        bands = self._recon_bands_check(bands)
        if max_orientations is not None:
            for i in bands:
                if i >= max_orientations:
                    warnings.warn(("You wanted band %d in the reconstruction but max_orientation"
                                   " is %d, so we're ignoring that band" % (i, max_orientations)))
            bands = [i for i in bands if i < max_orientations]
        recon_keys = []
        for level in levels:
            # residual highpass and lowpass
            if isinstance(level, str):
                recon_keys.append(level)
            # else we have to get each of the (specified) bands at
            # that level
            else:
                recon_keys.extend([(level, band) for band in bands])
        return recon_keys


class SteerablePyramidBase(Pyramid):
    """base class for steerable pyramid

    should not be called directly, we just use it so we can make both SteerablePyramidFreq and
    SteerablePyramidSpace inherit the steer_coeffs function

    """
    def __init__(self, image, edge_type):
        super().__init__(image=image, edge_type=edge_type)

    def steer_coeffs(self, angles, even_phase=True):
        """Steer pyramid coefficients to the specified angles

        This allows you to have filters that have the Gaussian derivative order specified in
        construction, but arbitrary angles or number of orientations.

        Parameters
        ----------
        angles : `list`
            list of angles (in radians) to steer the pyramid coefficients to
        even_phase : `bool`
            specifies whether the harmonics are cosine or sine phase aligned about those positions.

        Returns
        -------
        resteered_coeffs : `dict`
            dictionary of re-steered pyramid coefficients. will have the same number of scales as
            the original pyramid (though it will not contain the residual highpass or lowpass).
            like `self.pyr_coeffs`, keys are 2-tuples of ints indexing the scale and orientation,
            but now we're indexing `angles` instead of `self.num_orientations`.
        resteering_weights : `dict`
            dictionary of weights used to re-steer the pyramid coefficients. will have the same
            keys as `resteered_coeffs`.

        """
        resteered_coeffs = {}
        resteering_weights = {}
        for i in range(self.num_scales):
            basis = np.vstack([self.pyr_coeffs[(i, j)].flatten() for j in
                               range(self.num_orientations)]).T
            for j, a in enumerate(angles):
                res, steervect = steer(basis, a, return_weights=True, even_phase=even_phase)
                resteered_coeffs[(i, j)] = res.reshape(self.pyr_coeffs[(i, 0)].shape)
                resteering_weights[(i, j)] = steervect

        return resteered_coeffs, resteering_weights



from . import pyramids

from .pyramids.c.wrapper import corrDn, upConv, pointOp
from .pyramids.filters import named_filter, binomial_filter, steerable_filters

from .tools import synthetic_images
from .tools.convolutions import blurDn, blur, upBlur, image_gradient, rconv2
from .tools.display import imshow, animshow, pyrshow, make_figure
from .tools.image_stats import image_compare, image_stats, range, skew, var, entropy
from .tools.utils import rcosFn, matlab_histo, matlab_round, project_polar_to_cartesian
from .tools.compare_matpyrtools import comparePyr, compareRecon

from .version import version as __version__



#! /usr/bin/env python

from wheel.bdist_wheel import bdist_wheel
from setuptools import setup, Extension

# Adapted from the cibuildwheel example https://github.com/joerick/python-ctypes-package-sample
# it marks the wheel as not specific to the Python API version.
class WheelABINone(bdist_wheel):
    def finalize_options(self):
        bdist_wheel.finalize_options(self)
        self.root_is_pure = False

    def get_tag(self):
        _, _, plat = bdist_wheel.get_tag(self)
        return "py3", "none", plat


setup(
    ext_modules=[Extension('pyrtools.pyramids.c.wrapConv',
                           sources=['src/pyrtools/pyramids/c/py.c',
                                    'src/pyrtools/pyramids/c/convolve.c',
                                    'src/pyrtools/pyramids/c/edges.c',
                                    'src/pyrtools/pyramids/c/wrap.c',
                                    'src/pyrtools/pyramids/c/internal_pointOp.c'],
                           depends=['src/pyrtools/pyramids/c/meta.h',
                                    'src/pyrtools/pyramids/c/convolve.h',
                                    'src/pyrtools/pyramids/c/internal_pointOp.h'],
                           extra_compile_args=['-fPIC', '-shared'])],
    cmdclass={"bdist_wheel": WheelABINone},
    )



import numpy as np
from .GaussianPyramid import GaussianPyramid
from .filters import parse_filter
from .c.wrapper import upConv


class LaplacianPyramid(GaussianPyramid):
    """Laplacian pyramid

    Parameters
    ----------
    image : `array_like`
        1d or 2d image upon which to construct to the pyramid.
    height : 'auto' or `int`.
        The height of the pyramid. If 'auto', will automatically determine based on the size of
        `image`.
    downsample_filter_name : {'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3',
                              'daub4', 'qmf5', 'qmf9', 'qmf13'}
        name of filter to use for (separable) convolution to downsample the image. All scaled so
        L-2 norm is 1.0

        * `'binomN'` (default: 'binom5') - binomial coefficient filter of order N-1
        * `'haar'` - Haar wavelet
        * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
        * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
        * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_
    upsample_filter_name : {None, 'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3',
                            'daub4', 'qmf5', 'qmf9', 'qmf13'}
        name of filter to use as the "expansion" filter. All scaled so L-2 norm is 1.0

        * None (default) - same as `downsample_filter_name`
        * `'binomN'` - binomial coefficient filter of order N-1
        * `'haar'` - Haar wavelet
        * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
        * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
        * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_
    edge_type : {'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend', 'dont-compute'}
        Specifies how to handle edges. Options are:

        * `'circular'` - circular convolution
        * `'reflect1'` - reflect about the edge pixels
        * `'reflect2'` - reflect, doubling the edge pixels
        * `'repeat'` - repeat the edge pixels
        * `'zero'` - assume values of zero outside image boundary
        * `'extend'` - reflect and invert
        * `'dont-compute'` - zero output when filter overhangs imput boundaries.

    Attributes
    ----------
    image : `array_like`
        The input image used to construct the pyramid.
    image_size : `tuple`
        The size of the input image.
    pyr_type : `str` or `None`
        Human-readable string specifying the type of pyramid. For base class, is None.
    edge_type : `str`
        Specifies how edges were handled.
    pyr_coeffs : `dict`
        Dictionary containing the coefficients of the pyramid. Keys are `(level, band)` tuples and
        values are 1d or 2d numpy arrays (same number of dimensions as the input image)
    pyr_size : `dict`
        Dictionary containing the sizes of the pyramid coefficients. Keys are `(level, band)`
        tuples and values are tuples.
    is_complex : `bool`
        Whether the coefficients are complex- or real-valued. Only `SteerablePyramidFreq` can have
        a value of True, all others must be False.

    References
    ----------
    .. [1] J D Johnston, "A filter family designed for use in quadrature mirror filter banks",
       Proc. ICASSP, pp 291-294, 1980.
    .. [2] I Daubechies, "Orthonormal bases of compactly supported wavelets", Commun. Pure Appl.
       Math, vol. 42, pp 909-996, 1988.
    .. [3] E P Simoncelli,  "Orthogonal sub-band image transforms", PhD Thesis, MIT Dept. of Elec.
       Eng. and Comp. Sci. May 1988. Also available as: MIT Media Laboratory Vision and Modeling
       Technical Report #100.
    .. [4] E P Simoncelli and E H Adelson, "Subband image coding", Subband Transforms, chapter 4,
       ed. John W Woods, Kluwer Academic Publishers,  Norwell, MA, 1990, pp 143--192.

    """
    def __init__(self, image, height='auto', downsample_filter_name='binom5',
                 upsample_filter_name=None, edge_type='reflect1'):
        self.pyr_type = 'Laplacian'
        if upsample_filter_name is None:
            upsample_filter_name = downsample_filter_name
        super().__init__(image, height, downsample_filter_name, edge_type, upsample_filter_name=upsample_filter_name)


    def _build_pyr(self):
        """build the pyramid

        This should not be called directly by users, it's a helper function for constructing the
        pyramid

        """
        im = self.image
        for lev in range(self.num_scales - 1):
            im_next = self._build_next(im)
            im_recon = self._recon_prev(im_next, output_size=im.shape)
            im_residual = im - im_recon
            self.pyr_coeffs[(lev, 0)] = im_residual.copy()
            self.pyr_size[(lev, 0)] = im_residual.shape
            im = im_next
        self.pyr_coeffs[(lev+1, 0)] = im.copy()
        self.pyr_size[(lev+1, 0)] = im.shape


    def _recon_prev(self, image, output_size, upsample_filter=None, edge_type=None):
        """Reconstruct the previous level of the pyramid.

        Should not be called by users directly, this is a helper function for reconstructing the
        input image using pyramid coefficients.

        """
        if upsample_filter is None:
            upsample_filter = self.filters['upsample_filter']
        else:
            upsample_filter = parse_filter(upsample_filter, normalize=False)

        if edge_type is None:
            edge_type = self.edge_type

        if image.shape[0] == 1:
            res = upConv(image=image, filt=upsample_filter.T, edge_type=edge_type, step=(1, 2), stop=(output_size[0], output_size[1]))
        elif image.shape[1] == 1:
            res = upConv(image=image, filt=upsample_filter, edge_type=edge_type, step=(2, 1), stop=(output_size[0], output_size[1]))
        else:
            tmp = upConv(image=image, filt=upsample_filter, edge_type=edge_type, step=(2, 1), stop=(output_size[0], image.shape[1]))
            res = upConv(image=tmp, filt=upsample_filter.T, edge_type=edge_type, step=(1, 2), stop=(output_size[0], output_size[1]))
        return res


    def recon_pyr(self, upsample_filter_name=None, edge_type=None, levels='all'):
        """Reconstruct the input image using pyramid coefficients

        Parameters
        ----------
        upsample_filter_name : {None, 'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3',
                                'daub4', 'qmf5', 'qmf9', 'qmf13'}
            name of filter to use as "expansion" filter. All scaled so L-2 norm is 1.0

            * None (default) - use `self.upsample_filter_name`, the expansion filter set during
                               initialization.
            * `'binomN'` - binomial coefficient filter of order N-1
            * `'haar'` - Haar wavelet
            * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
            * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
            * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_
        edge_type : {None, 'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend',
                     'dont-compute'}
            Specifies how to handle edges. Options are:

            * None (default) - use `self.edge_type`, the edge_type used to construct the pyramid
            * `'circular'` - circular convolution
            * `'reflect1'` - reflect about the edge pixels
            * `'reflect2'` - reflect, doubling the edge pixels
            * `'repeat'` - repeat the edge pixels
            * `'zero'` - assume values of zero outside image boundary
            * `'extend'` - reflect and invert
            * `'dont-compute'` - zero output when filter overhangs imput boundaries.
        levels : `list`, `int`,  or {`'all'`, `'residual_highpass'`}
            If `list` should contain some subset of integers from `0` to `self.num_scales-1`
            (inclusive) and `'residual_lowpass'`. If `'all'`, returned value will contain all
            valid levels. Otherwise, must be one of the valid levels.

        Returns
        -------
        recon : `np.array`
            The reconstructed image.
        """
        recon_keys = self._recon_keys(levels, 'all')
        recon = np.zeros_like(self.pyr_coeffs[(self.num_scales-1, 0)])
        for lev in reversed(range(self.num_scales)):
            # upsample to generate higher reconolution image
            recon = self._recon_prev(recon, self.pyr_size[(lev, 0)], upsample_filter_name, edge_type)
            if (lev, 0) in recon_keys:
                recon += self.pyr_coeffs[(lev, 0)]
        return recon



import numpy as np
import warnings


def steer_to_harmonics_mtx(harmonics, angles=None, even_phase=True):
    '''Compute a steering matrix

    This maps a directional basis set onto the angular Fourier harmonics.

    Parameters
    ----------
    harmonics: `array_like`
        array specifying the angular harmonics contained in the steerable basis/filters.
    angles: `array_like` or None
        vector specifying the angular position of each filter (in radians). If None, defaults to
        `pi * np.arange(numh) / numh`, where `numh = harmonics.size + np.count_nonzero(harmonics)`
    even_phase : `bool`
        specifies whether the harmonics are cosine or sine phase aligned about those positions.

    Returns
    -------
    imtx : `np.array`
        This matrix is suitable for passing to the function `steer`.

    '''
    # default parameter
    numh = harmonics.size + np.count_nonzero(harmonics)
    if angles is None:
        angles = np.pi * np.arange(numh) / numh

    # Compute inverse matrix, which maps to Fourier components onto
    # steerable basis
    imtx = np.zeros((angles.size, numh))
    col = 0
    for h in harmonics:
        args = h * angles
        if h == 0:
            imtx[:, col] = np.ones(angles.shape)
            col += 1
        elif even_phase:
            imtx[:, col] = np.cos(args)
            imtx[:, col+1] = np.sin(args)
            col += 2
        else:  # odd phase
            imtx[:, col] = np.sin(args)
            imtx[:, col+1] = -1.0 * np.cos(args)
            col += 2

    r = np.linalg.matrix_rank(imtx)
    if r < np.min(imtx.shape):
        warnings.warn("Matrix is not full rank")

    return np.linalg.pinv(imtx)


def steer(basis, angle, harmonics=None, steermtx=None, return_weights=False, even_phase=True):
    '''Steer BASIS to the specfied ANGLE.

    Parameters
    ----------
    basis : `array_like`
        array whose columns are vectorized rotated copies of a steerable function, or the responses
        of a set of steerable filters.
    angle : `array_like` or `int`
        scalar or column vector the size of the basis. specifies the angle(s) (in radians) to
        steer to
    harmonics : `list` or None
        a list of harmonic numbers indicating the angular harmonic content of the basis. if None
        (default), N even or odd low frequencies, as for derivative filters
    steermtx : `array_like` or None.
        matrix which maps the filters onto Fourier series components (ordered [cos0 cos1 sin1 cos2
        sin2 ... sinN]). See steer_to_harmonics_mtx function for more details. If None (default),
        assumes cosine phase harmonic components, and filter positions at 2pi*n/N.
    return_weights : `bool`
        whether to return the weights or not.
    even_phase : `bool`
        specifies whether the harmonics are cosine or sine phase aligned about those positions.

    Returns
    -------
    res : `np.array`
        the resteered basis
    steervect : `np.array`
        the weights used to resteer the basis. only returned if `return_weights` is True
    '''

    num = basis.shape[1]

    if isinstance(angle, (int, float)):
        angle = np.asarray([angle])
    else:
        if angle.shape[0] != basis.shape[0] or angle.shape[1] != 1:
            raise Exception("""ANGLE must be a scalar, or a column vector
                                    the size of the basis elements""")

    # If HARMONICS is not specified, assume derivatives.
    if harmonics is None:
        harmonics = np.arange(1 - (num % 2), num, 2)

    if len(harmonics.shape) == 1 or harmonics.shape[0] == 1:
        # reshape to column matrix
        harmonics = harmonics.reshape(harmonics.shape[0], 1)
    elif harmonics.shape[0] != 1 and harmonics.shape[1] != 1:
        raise Exception('input parameter HARMONICS must be 1D!')

    if 2 * harmonics.shape[0] - (harmonics == 0).sum() != num:
        raise Exception('harmonics list is incompatible with basis size!')

    # If STEERMTX not passed, assume evenly distributed cosine-phase filters:
    if steermtx is None:
        steermtx = steer_to_harmonics_mtx(harmonics, np.pi * np.arange(num) / num,
                                          even_phase=even_phase)

    steervect = np.zeros((angle.shape[0], num))
    arg = angle * harmonics[np.nonzero(harmonics)[0]].T
    if all(harmonics):
        steervect[:, range(0, num, 2)] = np.cos(arg)
        steervect[:, range(1, num, 2)] = np.sin(arg)
    else:
        steervect[:, 0] = np.ones((arg.shape[0], 1))
        steervect[:, range(1, num, 2)] = np.cos(arg)
        steervect[:, range(2, num, 2)] = np.sin(arg)

    steervect = np.dot(steervect, steermtx)

    if steervect.shape[0] > 1:
        tmp = np.dot(basis, steervect)
        res = sum(tmp).T
    else:
        res = np.dot(basis, steervect.T)

    if return_weights:
        return res, np.asarray(steervect).reshape(num)
    else:
        return res
